from unittest import mock

from sqlmodel import Session, select

from murfey.util.db import (
    AutoProcProgram,
    ClassificationFeedbackParameters,
    DataCollection,
    DataCollectionGroup,
    ParticleSizes,
    ProcessingJob,
    TomogramPicks,
    TomographyProcessingParameters,
)
from murfey.workflows.tomo import picking
from tests.conftest import ExampleVisit, get_or_create_db_entry


def set_up_picking_db(murfey_db_session: Session):
    # Insert common elements needed in all picking tests
    dcg_entry: DataCollectionGroup = get_or_create_db_entry(
        murfey_db_session,
        DataCollectionGroup,
        lookup_kwargs={
            "id": 0,
            "session_id": ExampleVisit.murfey_session_id,
            "tag": "test_dcg",
        },
    )
    dc_entry: DataCollection = get_or_create_db_entry(
        murfey_db_session,
        DataCollection,
        lookup_kwargs={
            "id": 0,
            "tag": "test_dc",
            "dcg_id": dcg_entry.id,
        },
    )
    processing_job_entry: ProcessingJob = get_or_create_db_entry(
        murfey_db_session,
        ProcessingJob,
        lookup_kwargs={
            "id": 1,
            "recipe": "test_recipe",
            "dc_id": dc_entry.id,
        },
    )
    get_or_create_db_entry(
        murfey_db_session,
        AutoProcProgram,
        lookup_kwargs={
            "id": 0,
            "pj_id": processing_job_entry.id,
        },
    )
    get_or_create_db_entry(
        murfey_db_session,
        ClassificationFeedbackParameters,
        lookup_kwargs={
            "pj_id": processing_job_entry.id,
            "estimate_particle_diameter": True,
            "hold_class2d": False,
            "hold_class3d": False,
            "class_selection_score": 0,
            "star_combination_job": 0,
            "initial_model": "",
            "next_job": 0,
        },
    )
    return dcg_entry.id, dc_entry.id, processing_job_entry.id


def test_ids_tomo_classification(murfey_db_session: Session):
    dcg_id, first_dc_id, first_pj_id = set_up_picking_db(murfey_db_session)

    # Insert a second data collection, processing job and autoproc program
    second_dc: DataCollection = get_or_create_db_entry(
        murfey_db_session,
        DataCollection,
        lookup_kwargs={
            "id": 1,
            "tag": "second_dc",
            "dcg_id": dcg_id,
        },
    )
    second_pj: ProcessingJob = get_or_create_db_entry(
        murfey_db_session,
        ProcessingJob,
        lookup_kwargs={
            "id": 10,
            "recipe": "second_recipe",
            "dc_id": second_dc.id,
        },
    )
    get_or_create_db_entry(
        murfey_db_session,
        AutoProcProgram,
        lookup_kwargs={
            "id": 11,
            "pj_id": second_pj.id,
        },
    )

    returned_ids = picking._ids_tomo_classification(
        11, "test_recipe", murfey_db_session
    )
    assert returned_ids[0] == dcg_id
    assert returned_ids[1] == first_pj_id


@mock.patch("murfey.workflows.tomo.picking._transport_object")
@mock.patch("murfey.workflows.tomo.picking._ids_tomo_classification")
def test_picked_tomogram_not_run_class2d(
    mock_ids, mock_transport, murfey_db_session: Session, tmp_path
):
    """Run the picker feedback with less particles than needed for classification"""
    mock_ids.return_value = [2, 1]

    # Insert table dependencies
    set_up_picking_db(murfey_db_session)

    message = {
        "program_id": 0,
        "cbox_3d": f"{tmp_path}/AutoPick/job007/CBOX_3d/sample.cbox",
        "particle_count": 2,
        "particle_diameters": [10.1, 20.2],
        "pixel_size": 5.3,
        "register": "picked_tomogram",
        "tomogram": f"{tmp_path}/Tomograms/job006/tomograms/sample.mrc",
    }
    picking._register_picked_tomogram_use_diameter(message, murfey_db_session)

    mock_ids.assert_called_once_with(0, "em-tomo-class2d", murfey_db_session)

    tomograms_db = murfey_db_session.exec(
        select(TomogramPicks).where(TomogramPicks.pj_id == 1)
    ).one()
    assert tomograms_db.tomogram == message["tomogram"]
    assert tomograms_db.cbox_3d == message["cbox_3d"]
    assert tomograms_db.particle_count == 2
    assert tomograms_db.tomogram_pixel_size == 5.3

    added_picks = murfey_db_session.exec(
        select(ParticleSizes).where(ParticleSizes.pj_id == 1)
    ).all()
    assert len(added_picks) == 2
    assert added_picks[0].particle_size == 10.1
    assert added_picks[1].particle_size == 20.2

    mock_transport.send.assert_not_called()


@mock.patch("murfey.workflows.tomo.picking._transport_object")
@mock.patch("murfey.workflows.tomo.picking._ids_tomo_classification")
def test_picked_tomogram_run_class2d_with_diameter(
    mock_ids, mock_transport, murfey_db_session: Session, tmp_path
):
    """Run the picker feedback with a pre-determined particle diameter"""
    mock_transport.feedback_queue = "murfey_feedback"

    # Insert table dependencies
    dcg_id, dc_id, pj_id = set_up_picking_db(murfey_db_session)
    get_or_create_db_entry(
        murfey_db_session,
        TomographyProcessingParameters,
        lookup_kwargs={
            "dcg_id": dcg_id,
            "pixel_size": 1.34,
            "dose_per_frame": 1,
            "frame_count": 5,
            "tilt_axis": 0,
            "voltage": 300,
            "particle_diameter": 200,
        },
    )
    for particle in range(5001):
        get_or_create_db_entry(
            murfey_db_session,
            ParticleSizes,
            lookup_kwargs={
                "id": particle,
                "pj_id": pj_id,
                "particle_size": 100,
            },
        )

    mock_ids.return_value = [dcg_id, 1]

    message = {
        "session_id": 1,
        "program_id": 0,
        "cbox_3d": f"{tmp_path}/AutoPick/job007/CBOX_3d/sample.cbox",
        "particle_count": 2,
        "particle_diameters": [10.1, 20.2],
        "pixel_size": 5.3,
        "register": "picked_tomogram",
        "tomogram": f"{tmp_path}/Tomograms/job006/tomograms/sample.mrc",
    }
    picking._register_picked_tomogram_use_diameter(message, murfey_db_session)

    mock_ids.assert_called_once_with(0, "em-tomo-class2d", murfey_db_session)

    tomograms_db = murfey_db_session.exec(
        select(TomogramPicks).where(TomogramPicks.pj_id == 1)
    ).one()
    assert tomograms_db.tomogram == message["tomogram"]
    assert tomograms_db.cbox_3d == message["cbox_3d"]
    assert tomograms_db.particle_count == 2
    assert tomograms_db.tomogram_pixel_size == 5.3

    mock_transport.send.assert_called_once_with(
        "processing_recipe",
        {
            "parameters": {
                "tomogram": message["tomogram"],
                "cbox_3d": message["cbox_3d"],
                "pixel_size": message["pixel_size"],
                "particle_diameter": 200.0,
                "kv": 300,
                "node_creator_queue": "node_creator",
                "session_id": message["session_id"],
                "autoproc_program_id": 0,
                "batch_size": 5000,
                "nr_classes": 5,
                "picker_id": None,
                "class2d_grp_uuid": 6,
                "class_uuids": {str(i): i for i in range(1, 6)},
                "next_job": 9,
                "feedback_queue": "murfey_feedback",
            },
            "recipes": ["em-tomo-class2d"],
        },
        new_connection=True,
    )


@mock.patch("murfey.workflows.tomo.picking._transport_object")
@mock.patch("murfey.workflows.tomo.picking._ids_tomo_classification")
def test_picked_tomogram_run_class2d_estimate_diameter(
    mock_ids, mock_transport, murfey_db_session: Session, tmp_path
):
    """Run the picker feedback for Class2D, including diameter estimation"""
    mock_transport.feedback_queue = "murfey_feedback"

    # Insert table dependencies
    dcg_id, dc_id, pj_id = set_up_picking_db(murfey_db_session)
    get_or_create_db_entry(
        murfey_db_session,
        TomographyProcessingParameters,
        lookup_kwargs={
            "dcg_id": dcg_id,
            "pixel_size": 1.34,
            "dose_per_frame": 1,
            "frame_count": 5,
            "tilt_axis": 0,
            "voltage": 300,
            "particle_diameter": None,
        },
    )
    for particle in range(5001):
        get_or_create_db_entry(
            murfey_db_session,
            ParticleSizes,
            lookup_kwargs={
                "id": particle,
                "pj_id": pj_id,
                "particle_size": 100,
            },
        )
    # Insert one existing tomogram which should get flushed out
    get_or_create_db_entry(
        murfey_db_session,
        TomogramPicks,
        lookup_kwargs={
            "pj_id": pj_id,
            "tomogram": f"{tmp_path}/Tomograms/job006/tomograms/tomogram1.mrc",
            "cbox_3d": f"{tmp_path}/AutoPick/job007/CBOX_3d/tomogram1_picks.cbox",
            "particle_count": 10,
            "tomogram_pixel_size": 5.3,
        },
    )

    mock_ids.return_value = [dcg_id, 1]

    message = {
        "session_id": 1,
        "program_id": 0,
        "cbox_3d": f"{tmp_path}/AutoPick/job007/CBOX_3d/sample.cbox",
        "particle_count": 2,
        "particle_diameters": [10.1, 20.2],
        "pixel_size": 5.3,
        "register": "picked_tomogram",
        "tomogram": f"{tmp_path}/Tomograms/job006/tomograms/sample.mrc",
    }
    picking._register_picked_tomogram_use_diameter(message, murfey_db_session)

    mock_ids.assert_called_once_with(0, "em-tomo-class2d", murfey_db_session)

    # Two mock calls - one flushed tomogram and one new
    assert mock_transport.send.call_count == 2
    mock_transport.send.assert_any_call(
        "processing_recipe",
        {
            "parameters": {
                "tomogram": f"{tmp_path}/Tomograms/job006/tomograms/tomogram1.mrc",
                "cbox_3d": f"{tmp_path}/AutoPick/job007/CBOX_3d/tomogram1_picks.cbox",
                "pixel_size": 5.3,
                "particle_diameter": 100.0,
                "kv": 300,
                "node_creator_queue": "node_creator",
                "session_id": message["session_id"],
                "autoproc_program_id": 0,
                "batch_size": 5000,
                "nr_classes": 5,
                "picker_id": None,
                "class2d_grp_uuid": 12,
                "class_uuids": {str(i): i + 6 for i in range(1, 6)},
                "next_job": 9,
                "feedback_queue": "murfey_feedback",
            },
            "recipes": ["em-tomo-class2d"],
        },
        new_connection=True,
    )
    mock_transport.send.assert_any_call(
        "processing_recipe",
        {
            "parameters": {
                "tomogram": message["tomogram"],
                "cbox_3d": message["cbox_3d"],
                "pixel_size": message["pixel_size"],
                "particle_diameter": 100.0,
                "kv": 300,
                "node_creator_queue": "node_creator",
                "session_id": message["session_id"],
                "autoproc_program_id": 0,
                "batch_size": 5000,
                "nr_classes": 5,
                "picker_id": None,
                "class2d_grp_uuid": 18,
                "class_uuids": {str(i): i + 12 for i in range(1, 6)},
                "next_job": 11,
                "feedback_queue": "murfey_feedback",
            },
            "recipes": ["em-tomo-class2d"],
        },
        new_connection=True,
    )



import logging
from pathlib import Path
from typing import Dict, Optional

import xmltodict

from murfey.client.context import Context, ensure_dcg_exists
from murfey.client.contexts.spa import _file_transferred_to, _get_source
from murfey.client.instance_environment import MurfeyInstanceEnvironment
from murfey.util.client import capture_post
from murfey.util.spa_metadata import (
    FoilHoleInfo,
    get_grid_square_atlas_positions,
    grid_square_data,
)

logger = logging.getLogger("murfey.client.contexts.spa_metadata")


def _foil_hole_positions(xml_path: Path, grid_square: int) -> Dict[str, FoilHoleInfo]:
    with open(xml_path, "r") as xml:
        for_parsing = xml.read()
        data = xmltodict.parse(for_parsing)
    data = data["GridSquareXml"]
    if "TargetLocationsEfficient" in data["TargetLocations"].keys():
        # Grids with regular foil holes
        serialization_array = data["TargetLocations"]["TargetLocationsEfficient"][
            "a:m_serializationArray"
        ]
    elif "TargetLocations" in data["TargetLocations"].keys():
        # Lacey grids
        serialization_array = data["TargetLocations"]["TargetLocations"][
            "a:m_serializationArray"
        ]
    else:
        logger.warning(f"Target locations not found for {str(xml_path)}")
        return {}
    required_key = ""
    for key in serialization_array.keys():
        if key.startswith("b:KeyValuePairOfintTargetLocation"):
            required_key = key
            break
    if not required_key:
        logger.info(f"Required key not found for {str(xml_path)}")
        return {}
    foil_holes = {}
    for fh_block in serialization_array[required_key]:
        if fh_block["b:value"]["IsNearGridBar"] == "false":
            image_paths = list(
                (xml_path.parent.parent).glob(
                    f"Images-Disc*/GridSquare_{grid_square}/FoilHoles/FoilHole_{fh_block['b:key']}_*.jpg"
                )
            )
            image_paths.sort(key=lambda x: x.stat().st_ctime)
            image_path: str = str(image_paths[-1]) if image_paths else ""
            pix_loc = fh_block["b:value"]["PixelCenter"]
            stage = fh_block["b:value"]["StagePosition"]
            diameter = fh_block["b:value"]["PixelWidthHeight"]["c:width"]
            foil_holes[fh_block["b:key"]] = FoilHoleInfo(
                id=int(fh_block["b:key"]),
                grid_square_id=grid_square,
                x_location=int(float(pix_loc["c:x"])),
                y_location=int(float(pix_loc["c:y"])),
                x_stage_position=float(stage["c:X"]),
                y_stage_position=float(stage["c:Y"]),
                image=str(image_path),
                diameter=int(float(diameter)),
            )
    return foil_holes


class SPAMetadataContext(Context):
    def __init__(self, acquisition_software: str, basepath: Path, token: str):
        super().__init__("SPA_metadata", acquisition_software, token)
        self._basepath = basepath

    def post_transfer(
        self,
        transferred_file: Path,
        environment: Optional[MurfeyInstanceEnvironment] = None,
        **kwargs,
    ):
        super().post_transfer(
            transferred_file=transferred_file,
            environment=environment,
            **kwargs,
        )

        if transferred_file.name == "EpuSession.dm" and environment:
            logger.info("EPU session metadata found")
            with open(transferred_file, "r") as epu_xml:
                data = xmltodict.parse(epu_xml.read())
            windows_path = data["EpuSessionXml"]["Samples"]["_items"]["SampleXml"][0][
                "AtlasId"
            ]["#text"]
            logger.info(f"Windows path to atlas metadata found: {windows_path}")
            visit_index = windows_path.split("\\").index(environment.visit)
            partial_path = "/".join(windows_path.split("\\")[visit_index + 1 :])
            logger.info("Partial Linux path successfully constructed from Windows path")

            source = _get_source(transferred_file, environment)
            if not source:
                logger.warning(
                    f"Source could not be identified for {str(transferred_file)}"
                )
                return

            if source:
                gs_pix_positions = get_grid_square_atlas_positions(
                    source.parent / partial_path
                )
                for images_disc in list(source.glob("Images-Disc*")) or [
                    source / "Images-Disc1"
                ]:
                    # Do the dcg registration for every Images-Disc with this session file
                    dcg_tag = ensure_dcg_exists(
                        collection_type="spa",
                        metadata_source=images_disc,
                        environment=environment,
                        token=self._token,
                    )
                    for gs, pos_data in gs_pix_positions.items():
                        if pos_data:
                            capture_post(
                                base_url=str(environment.url.geturl()),
                                router_name="session_control.spa_router",
                                function_name="register_grid_square",
                                token=self._token,
                                session_id=environment.murfey_session,
                                gsid=int(gs),
                                data={
                                    "tag": dcg_tag,
                                    "x_location": pos_data[0],
                                    "y_location": pos_data[1],
                                    "x_stage_position": pos_data[2],
                                    "y_stage_position": pos_data[3],
                                    "width": pos_data[4],
                                    "height": pos_data[5],
                                    "angle": pos_data[6],
                                },
                            )

        elif (
            transferred_file.suffix == ".dm"
            and transferred_file.name.startswith("GridSquare")
            and environment
        ):
            # Make sure we have a data collection group before trying to register grid square
            source = _get_source(transferred_file, environment=environment)
            if source is None:
                return None
            for images_disc in list(source.glob("Images-Disc*")) or [
                source / "Images-Disc1"
            ]:
                ensure_dcg_exists(
                    collection_type="spa",
                    metadata_source=images_disc,
                    environment=environment,
                    token=self._token,
                )

            gs_name = int(transferred_file.stem.split("_")[1])
            logger.info(
                f"Collecting foil hole positions for {str(transferred_file)} and grid square {gs_name}"
            )
            fh_positions = _foil_hole_positions(transferred_file, gs_name)
            visitless_source_search_dir = "/".join(
                [part for part in source.parts if part != environment.visit]
            ).replace("//", "/")
            visitless_source_images_dirs = sorted(
                Path(visitless_source_search_dir).glob("Images-Disc*"),
                key=lambda x: x.stat().st_ctime,
            )
            if not visitless_source_images_dirs:
                logger.warning(
                    f"Cannot find Images-Disc* in {visitless_source_search_dir}"
                )
                visitless_source_images_dirs = [
                    Path(visitless_source_search_dir) / "Images-Disc1"
                ]
            visitless_source = str(visitless_source_images_dirs[-1])

            if fh_positions:
                gs_info = grid_square_data(
                    transferred_file,
                    gs_name,
                )
                image_path = (
                    _file_transferred_to(
                        environment, source, Path(gs_info.image), self._token
                    )
                    if gs_info.image
                    else ""
                )
                capture_post(
                    base_url=str(environment.url.geturl()),
                    router_name="session_control.spa_router",
                    function_name="register_grid_square",
                    token=self._token,
                    session_id=environment.murfey_session,
                    gsid=gs_name,
                    data={
                        "tag": visitless_source,
                        "readout_area_x": gs_info.readout_area_x,
                        "readout_area_y": gs_info.readout_area_y,
                        "thumbnail_size_x": gs_info.thumbnail_size_x,
                        "thumbnail_size_y": gs_info.thumbnail_size_y,
                        "pixel_size": gs_info.pixel_size,
                        "image": str(image_path),
                    },
                )

            for fh, fh_data in fh_positions.items():
                capture_post(
                    base_url=str(environment.url.geturl()),
                    router_name="session_control.spa_router",
                    function_name="register_foil_hole",
                    token=self._token,
                    session_id=environment.murfey_session,
                    gs_name=gs_name,
                    data={
                        "name": fh,
                        "x_location": fh_data.x_location,
                        "y_location": fh_data.y_location,
                        "x_stage_position": fh_data.x_stage_position,
                        "y_stage_position": fh_data.y_stage_position,
                        "readout_area_x": fh_data.readout_area_x,
                        "readout_area_y": fh_data.readout_area_y,
                        "thumbnail_size_x": fh_data.thumbnail_size_x,
                        "thumbnail_size_y": fh_data.thumbnail_size_y,
                        "pixel_size": fh_data.pixel_size,
                        "diameter": fh_data.diameter,
                        "tag": visitless_source,
                        "image": fh_data.image,
                    },
                )



from __future__ import annotations

import logging
from pathlib import Path

from murfey.client.instance_environment import (
    MurfeyInstanceEnvironment,
    global_env_lock,
)
from murfey.util import secure_path
from murfey.util.client import capture_get, capture_post

logger = logging.getLogger("murfey.client.destinations")


def find_longest_data_directory(
    match_path: Path, data_directories: list[str] | list[Path]
) -> tuple[Path | None, Path | None]:
    """
    Determine the longest path in the data_directories list
    which the match path is relative to
    """
    base_dir: Path | None = None
    mid_dir: Path | None = None
    for dd in data_directories:
        dd_base = Path(dd).absolute()
        if match_path.absolute().is_relative_to(dd_base) and len(dd_base.parts) > (
            len(base_dir.parts) if base_dir else 0
        ):
            base_dir = dd_base
            mid_dir = match_path.absolute().relative_to(Path(base_dir)).parent
    return base_dir, mid_dir


def determine_default_destination(
    visit: str,
    source: Path,
    destination: str,
    environment: MurfeyInstanceEnvironment,
    token: str,
    touch: bool = False,
    extra_directory: str = "",
    use_suggested_path: bool = True,
) -> str:
    if not destination or not visit:
        raise ValueError(f"No destination ({destination}) or visit ({visit}) supplied")
    machine_data = capture_get(
        base_url=str(environment.url.geturl()),
        router_name="session_control.router",
        function_name="machine_info_by_instrument",
        token=token,
        instrument_name=environment.instrument_name,
    ).json()
    base_path, mid_path = find_longest_data_directory(
        source, machine_data.get("data_directories", [])
    )
    if not base_path:
        raise ValueError(f"No data directory found for {source}")
    if source.absolute() == base_path.absolute():
        raise ValueError(
            f"Source is the same as the base path {secure_path(source.absolute())}"
        )

    _default = f"{destination}/{visit}/{source.name}"
    if use_suggested_path:
        with global_env_lock:
            if source.name == "Images-Disc1":
                source_name = source.parent.name
            elif source.name.startswith("Sample"):
                source_name = f"{source.parent.name}_{source.name}"
            else:
                source_name = source.name
            if environment.destination_registry.get(source_name):
                _default = environment.destination_registry[source_name]
            else:
                suggested_path_response = capture_post(
                    base_url=str(environment.url.geturl()),
                    router_name="file_io_instrument.router",
                    function_name="suggest_path",
                    token=token,
                    visit_name=visit,
                    session_id=environment.murfey_session,
                    data={
                        "base_path": f"{destination}/{visit}/raw",
                        "touch": touch,
                        "extra_directory": extra_directory,
                    },
                )
                if suggested_path_response is None:
                    raise RuntimeError("Murfey server is unreachable")
                _default = suggested_path_response.json().get("suggested_path")
                environment.destination_registry[source_name] = _default
    return (
        _default + f"/{extra_directory}"
        if not _default.endswith("/")
        else _default + f"{extra_directory}"
    )



from __future__ import annotations

import copy
import os
import socket
from functools import lru_cache
from importlib.metadata import entry_points
from pathlib import Path
from typing import Any, Literal, Optional

import yaml
from pydantic import BaseModel, ConfigDict, RootModel, ValidationInfo, field_validator
from pydantic_settings import BaseSettings


class MagnificationTable(RootModel[dict[int, float]]):
    pass


CALIBRATIONS_VALIDATION_SCHEMAS = {
    "magnification": MagnificationTable,
}


class MachineConfig(BaseModel):  # type: ignore
    """
    Keys that describe the type of workflow conducted on the client side, and how
    Murfey will handle its data transfer and processing
    """

    # General info --------------------------------------------------------------------
    display_name: str = ""
    instrument_name: str = ""
    instrument_type: str = ""  # For use with hierarchical config files
    image_path: Optional[Path] = None
    machine_override: str = ""

    # Hardware and software -----------------------------------------------------------
    camera: str = "FALCON"
    superres: bool = False
    calibrations: dict[str, Any] = {}
    acquisition_software: list[str] = []
    software_versions: dict[str, str] = {}
    software_settings_output_directories: dict[str, list[str]] = {}
    data_required_substrings: dict[str, dict[str, list[str]]] = {}

    # Client side directory setup -----------------------------------------------------
    data_directories: list[Path] = []
    create_directories: list[str] = ["atlas"]
    analyse_created_directories: list[str] = []
    gain_reference_directory: Optional[Path] = None
    eer_fractionation_file_template: str = ""

    # Data transfer setup -------------------------------------------------------------
    # General setup
    data_transfer_enabled: bool = True
    substrings_blacklist: dict[str, list[str]] = {
        "directories": [],
        "files": [],
    }

    # Rsync setup
    rsync_url: str = ""
    rsync_module: str = ""
    rsync_basepath: Optional[Path] = None
    allow_removal: bool = False

    # Upstream data download setup
    upstream_data_directories: dict[str, Path] = {}  # Previous sessions
    upstream_data_download_directory: Optional[Path] = None  # Set by microscope config
    upstream_data_search_strings: dict[str, list[str]] = {}  # For glob search
    upstream_data_tiff_locations: list[str] = ["processed"]  # Location of CLEM TIFFs

    # Data processing setup -----------------------------------------------------------
    # General processing setup
    processing_enabled: bool = True
    process_by_default: bool = True
    gain_directory_name: str = "processing"
    process_multiple_datasets: bool = True
    processed_directory_name: str = "processed"
    processed_extra_directory: str = ""
    recipes: dict[str, str] = {
        "em-spa-bfactor": "em-spa-bfactor",
        "em-spa-class2d": "em-spa-class2d",
        "em-spa-class3d": "em-spa-class3d",
        "em-spa-preprocess": "em-spa-preprocess",
        "em-spa-refine": "em-spa-refine",
        "em-tomo-preprocess": "em-tomo-preprocess",
        "em-tomo-align": "em-tomo-align",
    }

    # Particle picking setup
    default_model: Optional[Path] = None
    picking_model_search_directory: str = "processing"
    initial_model_search_directory: str = "processing/initial_model"

    # Data analysis plugins
    external_executables: dict[str, str] = {}
    external_executables_eer: dict[str, str] = {}
    external_environment: dict[str, str] = {}
    plugin_packages: dict[str, Path] = {}

    # Server and network setup --------------------------------------------------------
    # Configurations and URLs
    security_configuration_path: Optional[Path] = None
    murfey_url: str = "http://localhost:8000"
    frontend_url: str = "http://localhost:3000"
    instrument_server_url: str = "http://localhost:8001"
    smartem_api_url: str = ""

    # Messaging queues
    failure_queue: str = ""
    node_creator_queue: str = "node_creator"
    notifications_queue: str = "pato_notification"

    # Pydantic BaseModel settings
    model_config = ConfigDict(extra="allow")

    @field_validator("calibrations", mode="before")
    @classmethod
    def validate_calibration_data(
        cls, v: dict[str, dict[Any, Any]]
    ) -> dict[str, dict[Any, Any]]:
        # Pass the calibration dictionaries through their matching Pydantic models, if any are set
        if isinstance(v, dict):
            validated = {}
            for (
                key,
                value,
            ) in v.items():
                model_cls = CALIBRATIONS_VALIDATION_SCHEMAS.get(key)
                if model_cls:
                    try:
                        # Validate and store as a dict object with the corrected types
                        validated[key] = model_cls.model_validate(value).root
                    except Exception as e:
                        raise ValueError(f"Validation failed for key '{key}': {e}")
                else:
                    validated[key] = value
            return validated
        # Let it validate and fail as-is
        return v

    @field_validator("software_versions", mode="before")
    @classmethod
    def validate_software_versions(cls, v: dict[str, Any]) -> dict[str, str]:
        # Software versions should be numerical strings, even if they appear int- or float-like
        if isinstance(v, dict):
            validated = {key: str(value) for key, value in v.items()}
            return validated
        # Let it validate and fail as-is
        return v


@lru_cache(maxsize=1)
def machine_config_from_file(
    config_file_path: Path,
    instrument_name: str,
) -> dict[str, MachineConfig]:
    """
    Loads the machine config YAML file and constructs instrument-specific configs from
    a hierarchical set of dictionary key-value pairs. It will populate the keys listed
    in the general dictionary, then update the keys specified in the shared instrument
    dictionary, before finally updating the keys for that specific instrument.
    """

    def _recursive_update(base: dict[str, Any], new: dict[str, Any]):
        """
        Helper function to recursively update nested dictionaries.

        If the old and new values are both dicts, it will add the new keys and values
        to the existing dictionary recursively without overwriting entries.

        If the old and new values are both lists, it will extend the existing list.
        For all other values, it will overwrite the existing value with the new one.
        """
        for key, value in new.items():
            # If new values are dicts and dict values already exist, do recursive update
            if key in base and isinstance(base[key], dict) and isinstance(value, dict):
                base[key] = _recursive_update(base[key], value)
            # If new values are lists and a list already exists, extend the list
            elif (
                key in base and isinstance(base[key], list) and isinstance(value, list)
            ):
                base[key].extend(value)
            # Otherwise, overwrite/add values as normal
            else:
                base[key] = copy.deepcopy(value)
        return base

    # Load the dict from the file
    with open(config_file_path, "r") as config_stream:
        master_config: dict[str, Any] = yaml.safe_load(config_stream)

    # Construct requested machine configs from the YAML file
    all_machine_configs: dict[str, MachineConfig] = {}
    for i in sorted(master_config.keys()):
        # Skip reserved top-level keys
        if i in ("general", "clem", "fib", "tem"):
            continue
        # If instrument name is set, skip irrelevant configs
        if instrument_name and i != instrument_name:
            continue

        # Construct instrument config hierarchically
        config: dict[str, Any] = {}

        # Populate with general values
        general_config: dict[str, Any] = master_config.get("general", {})
        config = _recursive_update(config, general_config)

        # Populate with shared instrument values
        instrument_config: dict[str, Any] = master_config.get(i, {})
        instrument_shared_config: dict[str, Any] = master_config.get(
            str(instrument_config.get("instrument_type", "")).lower(), {}
        )
        config = _recursive_update(config, instrument_shared_config)

        # Insert instrument-specific values
        config = _recursive_update(config, instrument_config)

        # Add to master dictionary
        all_machine_configs[i] = MachineConfig(**config)

    return all_machine_configs


class Security(BaseModel):
    # Murfey database settings
    murfey_db_credentials: Path
    crypto_key: str
    db: str = "postgres"
    sqlalchemy_pooling: bool = True

    # ISPyB settings
    ispyb_credentials: Optional[Path] = None

    # Murfey server connection settings
    auth_url: str = ""
    auth_type: Literal["password", "cookie"] = "password"
    auth_algorithm: str = ""
    auth_key: str = ""
    cookie_key: str = ""
    instrument_auth_url: str = ""
    instrument_auth_type: Literal["token", ""] = "token"
    allow_user_token: bool = False  # TUI 'user' token support
    session_validation: str = ""
    session_token_timeout: Optional[int] = None
    allow_origins: list[str] = ["*"]

    # RabbitMQ settings
    rabbitmq_credentials: Path
    feedback_queue: str = "murfey_feedback"

    # Graylog settings
    graylog_host: str = ""
    graylog_port: Optional[int] = None

    model_config = ConfigDict()

    @field_validator("graylog_port")
    def check_port_present_if_host_is(
        cls, v: Optional[int], info: ValidationInfo, **kwargs
    ) -> Optional[int]:
        if info.data.get("graylog_host") and v is None:
            raise ValueError("The Graylog port must be set if the Graylog host is")
        return v


def security_from_file(config_file_path: Path) -> Security:
    with open(config_file_path, "r") as config_stream:
        config = yaml.safe_load(config_stream)
    return Security(**config)


class Settings(BaseSettings):
    murfey_machine_configuration: str = ""
    murfey_security_configuration: str = ""


settings = Settings()


@lru_cache()
def get_hostname():
    return socket.gethostname()


def get_microscope(machine_config: MachineConfig | None = None) -> str:
    if machine_config:
        microscope_name = machine_config.machine_override or os.getenv("BEAMLINE", "")
    else:
        microscope_name = os.getenv("BEAMLINE", "")
    return microscope_name


@lru_cache(maxsize=1)
def get_security_config() -> Security:
    if settings.murfey_security_configuration:
        return security_from_file(Path(settings.murfey_security_configuration))
    if settings.murfey_machine_configuration and os.getenv("BEAMLINE"):
        machine_config = get_machine_config(instrument_name=os.getenv("BEAMLINE"))[
            os.getenv("BEAMLINE", "")
        ]
        if machine_config.security_configuration_path:
            return security_from_file(machine_config.security_configuration_path)
    return Security(
        rabbitmq_credentials="",
        session_validation="",
        murfey_db_credentials="",
        crypto_key="",
        auth_key="",
        auth_algorithm="",
        sqlalchemy_pooling=True,
    )


@lru_cache(maxsize=1)
def get_machine_config(instrument_name: str = "") -> dict[str, MachineConfig]:
    # Create an empty machine config as a placeholder
    machine_configs = {instrument_name: MachineConfig()}
    if settings.murfey_machine_configuration:
        machine_configs = machine_config_from_file(
            Path(settings.murfey_machine_configuration), instrument_name
        )
    return machine_configs


def get_extended_machine_config(
    extension_name: str, instrument_name: str = ""
) -> Optional[BaseModel]:
    machine_config = get_machine_config(instrument_name=instrument_name).get(
        instrument_name or get_microscope()
    )
    if not machine_config:
        return None
    model = list(entry_points(group="murfey.config", name=extension_name))[0].load()
    data = getattr(machine_config, extension_name, {})
    return model(**data)



import json
import logging
import time
from datetime import datetime
from typing import Any
from unittest.mock import AsyncMock, MagicMock

import pytest
from pytest_mock import MockerFixture

import murfey
from murfey.server.api.logging import forward_logs


@pytest.mark.asyncio
async def test_forward_logs(
    mocker: MockerFixture,
):
    # Create example log messages
    message_list = [
        json.dumps(
            {
                "name": f"murfey.{module_name}",
                "msg": "Starting Murfey server version {murfey.__version__}, listening on 0.0.0.0:8000",
                "args": [],
                "levelname": levelname,
                "levelno": levelno,
                "pathname": f"{murfey.__file__}/{module_name}/__init__.py",
                "filename": "__init__.py",
                "module": "__init__",
                "exc_info": None,
                "exc_text": None,
                "stack_info": None,
                "lineno": 76,
                "funcName": f"start_{module_name}",
                "created": time.time(),
                "msecs": 930.0,
                "relativeCreated": 1379.8329830169678,
                "thread": time.time_ns(),
                "threadName": "MainThread",
                "processName": "MainProcess",
                "process": time.time_ns(),
                "message": f"Starting Murfey server version {murfey.__version__}, listening on 0.0.0.0:8000",
                "type": "log",
            }
        )
        for module_name, levelname, levelno in (
            ("module_1", "DEBUG", logging.DEBUG),
            ("module_2", "INFO", logging.INFO),
            ("module_3", "WARNING", logging.WARNING),
            ("module_4", "ERROR", logging.ERROR),
        )
    ]

    # Create a mock request to pass to the function
    mock_request = MagicMock()
    mock_request.json = AsyncMock(return_value=message_list)

    # Mock the logging module
    mock_logging = mocker.patch("murfey.server.api.logging.logging")

    # Mock the 'getLogger()' and 'handle()' functions
    mock_logger = MagicMock()
    mock_logger.handle.return_value = None
    mock_logging.getLogger.return_value = mock_logger

    # Run the function and check that the results are as expected
    await forward_logs(mock_request)

    # Check that the correct logger name was called.
    for i, message in enumerate(message_list):
        # Process the message as in the actual function
        log_data: dict[str, Any] = json.loads(message)
        logger_name = log_data["name"]
        log_data.pop("msecs", None)
        log_data.pop("relativeCreated", None)
        client_timestamp = log_data.pop("created", 0)
        if client_timestamp:
            log_data["client_time"] = datetime.fromtimestamp(
                client_timestamp
            ).isoformat()
        log_data["client_host"] = None  # No host, as function is being tested directly

        # Check that messages are unpacked and handled in sequence
        mock_logging.getLogger.call_args_list[i][0][0] == logger_name
        mock_logger.handle.call_args_list[i][0][0] == logging.makeLogRecord(log_data)

    # Check that 'handle' was called for each message
    assert mock_logger.handle.call_count == len(message_list)



from __future__ import annotations

import logging
from pathlib import Path
from typing import Any, Callable

from werkzeug.utils import secure_filename

logger = logging.getLogger("murfey.util")


def sanitise(in_string: str) -> str:
    return in_string.replace("\r\n", "").replace("\n", "")


def sanitise_path(in_path: Path) -> Path:
    return Path("/".join(secure_filename(p) for p in in_path.parts))


def sanitise_nonpath(in_string: str) -> str:
    for c in ("\r\n", "\n", "/", "\\", ":", ";"):
        in_string = in_string.replace(c, "")
    return in_string


def secure_path(in_path: Path, keep_spaces: bool = False) -> Path:
    secured_parts = []
    for p, part in enumerate(in_path.parts):
        if p == 0 and ":" in part:
            secured_parts.append(secure_filename(part) + ":")
        elif " " in part and keep_spaces:
            secured_parts.append(part)
        elif part.endswith("_"):
            # Preserve all trailing underscores
            num_underscores = len(part) - len(part.rstrip("_"))
            secured_parts.append(secure_filename(part) + (num_underscores * "_"))
        else:
            secured_parts.append(secure_filename(part))
    return Path("/".join(secured_parts))


def posix_path(path: Path) -> str:
    """
    Converts a Windows-style path into a Posix one. Used primarily when running
    subproceses in bash terminals on Windows devices, which can only accept
    Posix paths.

    Returns it as a string because this path won't be recognised as an existing
    path when converted to a Path object.
    """
    path_parts = list(path.parts)
    # Check if it's a Windows-style path and converts it to a Posix one
    #   e.g.: C:\Users\user -> /c/Users/user
    if path_parts[0].endswith((":/", ":\\")):
        path_parts[0] = "/" + path_parts[0].strip(":/\\").lower()
        posix_path = "/".join(path_parts)
        return posix_path
    return str(path)


def safe_run(
    func: Callable,
    args: list | tuple = [],
    kwargs: dict[str, Any] = {},
    label: str = "",
):
    """
    A wrapper to encase individual functions in try-except blocks so that a warning
    is raised if the function fails, but the process continues as normal otherwise.
    """
    try:
        return func(*args, **kwargs)
    except Exception:
        logger.warning(
            f"Function {func.__name__!r} failed to run for object {label!r}",
            exc_info=True,
        )
        return None



import argparse
import asyncio
import json
from datetime import datetime
from functools import partial
from inspect import getfullargspec, iscoroutinefunction
from pathlib import Path
from queue import Empty, Queue

from sqlmodel import Session, create_engine
from workflows.transport.pika_transport import PikaTransport

from murfey.server.murfey_db import url
from murfey.server.run import _set_up_transport
from murfey.util.config import security_from_file


def dlq_purge(
    dlq_dump_path: Path, queue: str, rabbitmq_credentials: Path
) -> list[Path]:
    transport = PikaTransport()
    transport.load_configuration_file(rabbitmq_credentials)
    transport.connect()

    queue_to_purge = f"dlq.{queue}"
    idlequeue: Queue = Queue()
    exported_messages = []

    def receive_dlq_message(header: dict, message: dict) -> None:
        idlequeue.put_nowait("start")
        header["x-death"][0]["time"] = datetime.timestamp(header["x-death"][0]["time"])
        filename = dlq_dump_path / f"{queue}-{header['message-id']}"
        dlqmsg = {"header": header, "message": message}
        with filename.open("w") as fh:
            json.dump(dlqmsg, fh, indent=2, sort_keys=True)
        print(f"Message {header['message-id']} exported to {filename}")
        exported_messages.append(filename)
        transport.ack(header)
        idlequeue.put_nowait("done")

    print("Looking for DLQ messages in " + queue_to_purge)
    transport.subscribe(
        queue_to_purge,
        partial(receive_dlq_message),
        acknowledgement=True,
    )
    try:
        while True:
            idlequeue.get(True, 0.1)
    except Empty:
        print("Done dlq purge")
    transport.disconnect()
    return exported_messages


def handle_dlq_messages(messages_path: list[Path], rabbitmq_credentials: Path):
    transport = PikaTransport()
    transport.load_configuration_file(rabbitmq_credentials)
    transport.connect()

    for f, dlqfile in enumerate(messages_path):
        if not dlqfile.is_file():
            continue
        with open(dlqfile) as fh:
            dlqmsg = json.load(fh)
        header = dlqmsg["header"]
        header["dlq-reinjected"] = "True"

        drop_keys = {
            "message-id",
            "routing_key",
            "redelivered",
            "exchange",
            "consumer_tag",
            "delivery_mode",
        }
        clean_header = {k: str(v) for k, v in header.items() if k not in drop_keys}

        destination = header.get("x-death", [{}])[0].get("queue")
        transport.send(
            destination,
            dlqmsg["message"],
            headers=clean_header,
        )
        dlqfile.unlink()
        print(f"Reinjected {dlqfile}\n")

    transport.disconnect()


def handle_failed_posts(messages_path: list[Path], murfey_db: Session):
    """Deal with any messages that have been sent as failed client posts"""
    # These imports need to happen after transport object is configured
    import murfey.server.api.auth
    import murfey.server.api.bootstrap
    import murfey.server.api.clem
    import murfey.server.api.display
    import murfey.server.api.file_io_frontend
    import murfey.server.api.file_io_instrument
    import murfey.server.api.hub
    import murfey.server.api.instrument
    import murfey.server.api.mag_table
    import murfey.server.api.processing_parameters
    import murfey.server.api.prometheus
    import murfey.server.api.session_control
    import murfey.server.api.session_info
    import murfey.server.api.websocket
    import murfey.server.api.workflow

    for json_file in messages_path:
        with open(json_file, "r") as json_data:
            message = json.load(json_data)
        router_name = message.get("message", {}).get("router_name", "")
        router_base = router_name.split(".")[0]
        function_name = message.get("message", {}).get("function_name", "")
        if not router_name or not function_name:
            print(
                f"Cannot repost {json_file} as it does not have a router or function name"
            )
            continue

        try:
            function_to_call = getattr(
                getattr(murfey.server.api, router_base), function_name
            )
        except AttributeError:
            print(f"Cannot repost {json_file} as {function_name} does not exist")
            continue
        expected_args = getfullargspec(function_to_call)

        call_kwargs = message.get("message", {}).get("kwargs", {})
        call_data = message.get("message", {}).get("data", {})
        function_call_dict = {}

        try:
            for call_arg in expected_args.args:
                call_arg_type = expected_args.annotations.get(call_arg, str)
                if call_arg in call_kwargs.keys():
                    function_call_dict[call_arg] = call_arg_type(call_kwargs[call_arg])
                elif call_arg == "db":
                    function_call_dict["db"] = murfey_db
                else:
                    print(call_data, call_arg_type, call_arg)
                    function_call_dict[call_arg] = call_arg_type(**call_data)
        except TypeError as e:
            print(f"Cannot repost {json_file} due to argument error: {e}")
            continue

        try:
            if iscoroutinefunction(function_to_call):
                asyncio.run(function_to_call(**function_call_dict))
            else:
                function_to_call(**function_call_dict)
            print(f"Reposted {json_file}")
            json_file.unlink()
        except Exception as e:
            print(f"Failed to post {json_file} to {function_name}: {e}")


def run():
    """
    Method of checking and purging murfey queues on rabbitmq
    Two types of messages are possible:
    - failed client posts which need reposting to the murfey server API
    - feedback messages that can be sent back to rabbitmq
    """
    parser = argparse.ArgumentParser(
        description=(
            "Purge and reinject failed murfey messages. "
            "Provide security configuration and set machine configuration."
        )
    )
    parser.add_argument(
        "-c",
        "--config",
        help="Security config file",
        required=True,
    )
    parser.add_argument(
        "-d", "--dir", default="DLQ", help="Directory to export messages to"
    )
    args = parser.parse_args()

    # Read the security config file
    security_config = security_from_file(args.config)

    # Configure the transport
    PikaTransport().load_configuration_file(security_config.rabbitmq_credentials)
    _set_up_transport("PikaTransport")

    # Now import transport object which was set up in the above step
    from murfey.server import _transport_object

    _transport_object.feedback_queue = security_config.feedback_queue

    # Purge the queue and repost/reinject any messages found
    dlq_dump_path = Path(args.dir)
    dlq_dump_path.mkdir(parents=True, exist_ok=True)
    exported_messages = dlq_purge(
        dlq_dump_path,
        security_config.feedback_queue,
        security_config.rabbitmq_credentials,
    )

    # Set up database and retry api calls
    _url = url(security_config)
    engine = create_engine(_url)
    with Session(engine) as murfey_db:
        handle_failed_posts(exported_messages, murfey_db)

    # Reinject all remaining messages to rabbitmq
    handle_dlq_messages(exported_messages, security_config.rabbitmq_credentials)

    # Clean up any created directories
    try:
        dlq_dump_path.rmdir()
    except OSError:
        print(f"Cannot remove {dlq_dump_path} as it is not empty")
    print("Done")



import logging
from pathlib import Path
from typing import Optional

import xmltodict

from murfey.client.context import Context, _atlas_destination
from murfey.client.contexts.spa import _get_source
from murfey.client.instance_environment import MurfeyInstanceEnvironment
from murfey.util.client import capture_post

logger = logging.getLogger("murfey.client.contexts.atlas")


class AtlasContext(Context):
    def __init__(self, acquisition_software: str, basepath: Path, token: str):
        super().__init__("Atlas", acquisition_software, token)
        self._basepath = basepath

    def post_transfer(
        self,
        transferred_file: Path,
        environment: Optional[MurfeyInstanceEnvironment] = None,
        **kwargs,
    ):
        super().post_transfer(
            transferred_file=transferred_file,
            environment=environment,
            **kwargs,
        )
        if self._acquisition_software == "serialem":
            self.post_transfer_serialem(
                transferred_file, environment=environment, **kwargs
            )
        else:
            self.post_transfer_epu(transferred_file, environment=environment, **kwargs)

    def post_transfer_serialem(
        self,
        transferred_file: Path,
        environment: Optional[MurfeyInstanceEnvironment] = None,
        **kwargs,
    ):
        if environment and transferred_file.suffix == ".mrc":
            source = _get_source(transferred_file, environment)
            if source:
                capture_post(
                    base_url=str(environment.url.geturl()),
                    router_name="session_control.spa_router",
                    function_name="register_atlas",
                    token=self._token,
                    session_id=environment.murfey_session,
                    data={
                        "name": transferred_file.stem,
                        "acquisition_uuid": environment.acquisition_uuid,
                    },
                )

    def post_transfer_epu(
        self,
        transferred_file: Path,
        environment: Optional[MurfeyInstanceEnvironment] = None,
        **kwargs,
    ):
        if (
            environment
            and "Atlas_" in transferred_file.stem
            and transferred_file.suffix == ".mrc"
        ):
            source = _get_source(transferred_file, environment)
            if source:
                transferred_atlas_name = _atlas_destination(
                    environment, source, self._token
                ) / transferred_file.relative_to(source.parent)
                capture_post(
                    base_url=str(environment.url.geturl()),
                    router_name="session_control.spa_router",
                    function_name="make_atlas_jpg",
                    token=self._token,
                    session_id=environment.murfey_session,
                    data={"path": str(transferred_atlas_name).replace("//", "/")},
                )
                logger.info(
                    f"Submitted request to create JPG image of atlas {str(transferred_atlas_name)!r}"
                )
        elif (
            environment
            and "Atlas_" in transferred_file.stem
            and transferred_file.suffix == ".xml"
        ):
            source = _get_source(transferred_file, environment)
            if source:
                atlas_mrc = transferred_file.with_suffix(".mrc")
                transferred_atlas_jpg = _atlas_destination(
                    environment, source, self._token
                ) / atlas_mrc.relative_to(source.parent).with_suffix(".jpg")

                with open(transferred_file, "rb") as atlas_xml:
                    atlas_xml_data = xmltodict.parse(atlas_xml)
                    atlas_original_pixel_size = float(
                        atlas_xml_data["MicroscopeImage"]["SpatialScale"]["pixelSize"][
                            "x"
                        ]["numericValue"]
                    )

                # need to calculate the pixel size of the downscaled image
                atlas_pixel_size = atlas_original_pixel_size * 7.8

                for p in transferred_file.parts:
                    if p.startswith("Sample"):
                        sample = int(p.replace("Sample", ""))
                        break
                else:
                    logger.warning(
                        f"Sample could not be identified for {transferred_file}"
                    )
                    return

                dcg_data = {
                    "experiment_type_id": 44,  # Atlas
                    "tag": str(transferred_file.parent),
                    "atlas": str(transferred_atlas_jpg).replace("//", "/"),
                    "sample": sample,
                    "atlas_pixel_size": atlas_pixel_size,
                }
                capture_post(
                    base_url=str(environment.url.geturl()),
                    router_name="workflow.router",
                    function_name="register_dc_group",
                    token=self._token,
                    visit_name=environment.visit,
                    session_id=environment.murfey_session,
                    data=dcg_data,
                )
                logger.info(
                    f"Registered data collection group for atlas {str(transferred_atlas_jpg)!r}"
                )



import json
import os
import subprocess
import sys
from pathlib import Path
from queue import Empty
from unittest import mock

from murfey.cli import repost_failed_calls
from murfey.util.config import security_from_file
from murfey.util.db import Tilt


@mock.patch("murfey.cli.repost_failed_calls.PikaTransport")
@mock.patch("murfey.cli.repost_failed_calls.Queue")
def test_dlq_purge(mock_queue, mock_transport, tmp_path):
    """Test the dlq purging function.
    Currently doesn't test saving the message, as the subscribe is mocked out"""
    mock_queue().get.return_value = {"message": "dummy"}
    mock_queue().get.side_effect = [None, Empty]

    exported_messages = repost_failed_calls.dlq_purge(
        tmp_path / "DLQ", "dummy", tmp_path / "config_file"
    )

    # The transport should be connected to and subscribes to the queue
    mock_transport.assert_called_once()
    mock_transport().load_configuration_file.assert_called_with(
        tmp_path / "config_file"
    )
    mock_transport().connect.assert_called_once()
    mock_transport().subscribe.assert_called_with(
        "dlq.dummy",
        mock.ANY,
        acknowledgement=True,
    )
    mock_transport().disconnect.assert_called_once()

    # Should read from the queue
    mock_queue().get.assert_any_call(True, 0.1)

    # Ideally this test would return the message, but the partial isn't called yet
    assert exported_messages == []


@mock.patch("murfey.cli.repost_failed_calls.PikaTransport")
def test_handle_dlq_messages(mock_transport, tmp_path):
    """Reinject some example messages"""
    # Create two sample messages
    messages_paths_list: list[Path] = [tmp_path / "not_a_message"]
    messages_dict: dict[str, dict] = {
        "msg1": {
            "header": {
                "x-death": [{"queue": "queue_msg1"}],
                "message-id": 1,
                "routing_key": "dlq.queue_msg1",
                "redelivered": True,
                "exchange": "",
                "consumer_tag": "1",
                "delivery_mode": 2,
                "other_key": "value",
            },
            "message": {"parameters": "msg1"},
        },
        "msg2": {
            "header": {"x-death": [{"queue": "queue_msg2"}]},
            "message": {"content": "msg2"},
        },
    }
    for file_name, message in messages_dict.items():
        messages_paths_list.append(tmp_path / file_name)
        with open(tmp_path / file_name, "w") as msg_file:
            json.dump(message, msg_file)

    # Send the two messages, plus a file that is not a message
    repost_failed_calls.handle_dlq_messages(
        messages_path=messages_paths_list,
        rabbitmq_credentials=tmp_path / "config_file",
    )

    mock_transport.assert_called_once()
    mock_transport().load_configuration_file.assert_called_with(
        tmp_path / "config_file"
    )
    mock_transport().connect.assert_called_once()

    # Only two messages should have been sent, the rest are invalid so are skipped
    assert mock_transport().send.call_count == 2
    mock_transport().send.assert_any_call(
        "queue_msg1",
        {"parameters": "msg1"},
        headers={
            "x-death": "[{'queue': 'queue_msg1'}]",
            "other_key": "value",
            "dlq-reinjected": "True",
        },
    )
    mock_transport().send.assert_any_call(
        "queue_msg2",
        {"content": "msg2"},
        headers={"x-death": "[{'queue': 'queue_msg2'}]", "dlq-reinjected": "True"},
    )

    # Removal and waiting
    assert not (tmp_path / "msg1").is_file()
    assert not (tmp_path / "msg2").is_file()
    mock_transport().disconnect.assert_called_once()


def test_handle_failed_posts(tmp_path):
    """Test that the API is called with any failed client post messages"""
    # Create some sample messages
    messages_paths_list: list[Path] = []
    messages_dict: dict[str, dict] = {
        "msg1": {
            "message": {
                "router_name": "workflow.tomo_router",
                "function_name": "register_completed_tilt_series",
                "kwargs": {"visit_name": "cm12345-1", "session_id": 1},
                "data": {
                    "tags": ["tag"],
                    "source": "source",
                    "tilt_series_lengths": [10],
                },
            },
        },  # normal example
        "msg2": {
            "message": {
                "router_name": "workflow.tomo_router",
                "function_name": "register_tilt",
                "kwargs": {"visit_name": "cm12345-1", "session_id": 1},
                "data": {
                    "tilt_series_tag": "tag",
                    "source": "source",
                    "movie_path": "path",
                },
            },
        },  # async example
        "msg3": {
            "message": {
                "router_name": "workflow.tomo_router",
                "function_name": "register_completed_tilt_series",
                "data": {"tags": ["tag"]},
            }
        },
        "msg4": {
            "message": {"function_name": "dummy"},  # does not have a router
        },
        "msg5": {
            "message": {"router_name": "workflow"},  # does not have a function
        },
        "msg6": {
            "message": {
                "router_name": "workflow",
                "function_name": "dummy",
            },  # function does not exist
        },
    }
    for file_name, message in messages_dict.items():
        messages_paths_list.append(tmp_path / file_name)
        with open(tmp_path / file_name, "w") as msg_file:
            json.dump(message, msg_file)

    mock_db = mock.Mock()
    mock_exec_return = mock.Mock()
    mock_exec_return.all.return_value = []
    mock_db.exec.return_value = mock_exec_return
    repost_failed_calls.handle_failed_posts(messages_paths_list, mock_db)

    # Check the failed posts were resent
    assert mock_db.exec.call_count == 3
    assert mock_db.exec().one.call_count == 1
    assert mock_db.exec().all.call_count == 2
    assert mock_exec_return.one.call_count == 1
    assert mock_exec_return.all.call_count == 2
    assert mock_db.commit.call_count == 3
    mock_db.add.assert_called_once_with(
        Tilt(movie_path="path", tilt_series_id=mock.ANY, motion_corrected=False)
    )

    # Check only the failed post which was successfully reinjected got deleted
    assert not (tmp_path / "msg1").is_file()  # got resent
    assert not (tmp_path / "msg2").is_file()  # got resent
    assert (tmp_path / "msg3").is_file()  # failed reinjection
    assert (tmp_path / "msg4").is_file()  # does not have a router
    assert (tmp_path / "msg5").is_file()  # does not have a function
    assert (tmp_path / "msg6").is_file()  # function does not exist


@mock.patch("workflows.transport.pika_transport.PikaTransport")
@mock.patch("murfey.cli.repost_failed_calls.PikaTransport")
@mock.patch("murfey.cli.repost_failed_calls.dlq_purge")
@mock.patch("murfey.cli.repost_failed_calls.handle_failed_posts")
@mock.patch("murfey.cli.repost_failed_calls.handle_dlq_messages")
@mock.patch("murfey.cli.repost_failed_calls.url")
@mock.patch("murfey.cli.repost_failed_calls.create_engine")
@mock.patch("murfey.cli.repost_failed_calls.Session")
def test_run_repost_failed_calls(
    mock_db_session,
    mock_db_engine,
    mock_db_url,
    mock_reinject,
    mock_repost,
    mock_purge,
    mock_pika,
    mock_workflows,
    mock_security_configuration,
):
    mock_session = mock.MagicMock()

    mock_db_url.return_value = "db_url"
    mock_db_engine.return_value = "db_engine"
    mock_db_session.return_value = mock_session
    mock_purge.return_value = ["/path/to/msg1"]

    config_file = mock_security_configuration
    with open(config_file) as f:
        security_config_dict = json.load(f)

    sys.argv = [
        "murfey.repost_failed_calls",
        "--config",
        str(config_file),
        "--dir",
        "DLQ_dir",
    ]
    repost_failed_calls.run()

    security_config_class = security_from_file(config_file)
    mock_db_url.assert_called_with(security_config_class)
    mock_db_engine.assert_called_with("db_url")
    mock_db_session.assert_called_with("db_engine")

    mock_pika().load_configuration_file.assert_called_with(
        Path("/path/to/rabbitmq.yaml")
    )
    mock_workflows().connect.assert_called_once()

    mock_purge.assert_called_once_with(
        Path("DLQ_dir"),
        "murfey_feedback",
        Path(security_config_dict["rabbitmq_credentials"]),
    )
    mock_repost.assert_called_once_with(["/path/to/msg1"], mock_session.__enter__())
    mock_reinject.assert_called_once_with(
        ["/path/to/msg1"], Path(security_config_dict["rabbitmq_credentials"])
    )


def test_repost_failed_calls_exists(mock_security_configuration):
    """Test the CLI is made"""
    os.environ["MURFEY_SECURITY_CONFIGURATION"] = str(mock_security_configuration)
    result = subprocess.run(
        [
            "murfey.repost_failed_calls",
            "--help",
        ],
        capture_output=True,
    )
    assert not result.returncode

    # Find the first line of the help and strip out all the spaces and newlines
    stdout_as_string = result.stdout.decode("utf8", "replace")
    cleaned_help_line = (
        stdout_as_string.split("\n\n")[0].replace("\n", "").replace(" ", "")
    )
    assert cleaned_help_line == ("usage:murfey.repost_failed_calls[-h]-cCONFIG[-dDIR]")



from unittest.mock import MagicMock

import pytest
from pytest_mock import MockerFixture

from murfey.workflows.register_processing_job import run

register_processing_job_params_matrix = [
    # ISPyB session present | DC search result | PJ search result | APP search result | Insert ISPyB job | Update processing status
    (v0, v1, v2, v3, v4, v5)
    for v0 in (0, None)
    for v1 in (0, None)
    for v2 in (0, None)
    for v3 in (0, None)
    for v4 in (0, None)
    for v5 in (0, None)
]


@pytest.mark.parametrize("test_params", register_processing_job_params_matrix)
def test_run(
    mocker: MockerFixture,
    test_params: tuple[
        int | None, int | None, int | None, int | None, int | None, int | None
    ],
):
    # Unpack test params
    ispyb_session, dc_result, pj_result, app_result, insert_job, update_status = (
        test_params
    )

    # Create mocks
    # Transport object functions
    mock_transport_object = mocker.patch(
        "murfey.workflows.register_processing_job._transport_object"
    )
    mock_transport_object.do_create_ispyb_job.return_value = {
        "return_value": insert_job
    }
    mock_transport_object.do_update_processing_status.return_value = {
        "return_value": update_status
    }

    # ISPyB session
    mock_ispyb_session = mocker.patch(
        "murfey.workflows.register_processing_job.ISPyBSession"
    )
    mock_ispyb_session.return_value = ispyb_session

    # Murfey database
    mock_murfey_dc = MagicMock()
    mock_murfey_dc.id = dc_result
    mock_murfey_pj = MagicMock()
    mock_murfey_pj.id = pj_result
    mock_murfey_app = MagicMock()
    mock_murfey_app.id = app_result

    # Set up side effects depending on route taken through the function
    db_call_order = [[[mock_murfey_dc]] if dc_result is not None else []]
    if dc_result is not None:
        db_call_order.append([mock_murfey_pj] if pj_result is not None else [])
    if pj_result is not None or insert_job is not None or ispyb_session is None:
        db_call_order.append([mock_murfey_app] if app_result is not None else [])
    mock_murfey_db = MagicMock()
    mock_murfey_db.exec.return_value.all.side_effect = db_call_order

    # Mock Prometheus object
    mock_prom = mocker.patch("murfey.workflows.register_processing_job.prom")

    # Run function and check results and calls
    message = {
        "session_id": 0,
        "source": "some_path",
        "tag": "some_tag",
        "recipe": "some_recipe",
        "parameters": {},
        "job_parameters": {"dummy": "dummy"},
    }
    result = run(message=message, murfey_db=mock_murfey_db)
    if dc_result is not None:
        if pj_result is not None:
            mock_prom.preprocessed_movies.labels.assert_called_once()
            if app_result is not None:
                assert {"success": True}
            else:
                if update_status is not None:
                    assert result == {"success": True}
                else:
                    if ispyb_session is not None:
                        assert result == {"success": False, "requeue": True}
                    else:
                        assert result == {"success": True}
        else:
            if ispyb_session is not None:
                mock_transport_object.do_create_ispyb_job.assert_called_once()
                if insert_job is not None:
                    if app_result is not None:
                        assert result == {"success": True}
                    else:
                        if update_status is not None:
                            assert result == {"success": True}
                        else:
                            assert result == {"success": False, "requeue": True}
                else:
                    assert result == {"success": False, "requeue": True}
            else:
                assert result == {"success": True}
    else:
        assert result == {"success": False, "requeue": True}



from __future__ import annotations

import json
import logging
from datetime import datetime
from typing import Any, TypeVar

from fastapi import APIRouter, WebSocket, WebSocketDisconnect
from sqlmodel import Session, select

from murfey.server.murfey_db import get_murfey_db_session
from murfey.util import sanitise
from murfey.util.db import ClientEnvironment

T = TypeVar("T")

ws = APIRouter(prefix="/ws", tags=["Websocket"])
log = logging.getLogger("murfey.server.api.websocket")


class ConnectionManager:
    def __init__(self):
        self.active_connections: dict[int | str, WebSocket] = {}

    async def connect(
        self,
        websocket: WebSocket,
        client_id: int | str,
        register_client: bool = True,
    ):
        await websocket.accept()
        self.active_connections[client_id] = websocket
        if register_client:
            if not isinstance(client_id, int):
                raise ValueError(
                    "To register a client the client ID must be an integer"
                )
            self._register_new_client(client_id)

    @staticmethod
    def _register_new_client(client_id: int):
        log.debug(f"Registering new client with ID {client_id}")
        new_client = ClientEnvironment(client_id=client_id, connected=True)
        murfey_db: Session = next(get_murfey_db_session())
        murfey_db.add(new_client)
        murfey_db.commit()
        murfey_db.close()

    def disconnect(self, client_id: int | str, unregister_client: bool = True):
        self.active_connections.pop(client_id)
        if unregister_client:
            murfey_db: Session = next(get_murfey_db_session())
            client_env = murfey_db.exec(
                select(ClientEnvironment).where(
                    ClientEnvironment.client_id == client_id
                )
            ).one()
            murfey_db.delete(client_env)
            murfey_db.commit()
            murfey_db.close()

    async def broadcast(self, message: str):
        for connection in self.active_connections:
            await self.active_connections[connection].send_text(message)


manager = ConnectionManager()


@ws.websocket("/connect/{client_id}")
async def websocket_connection_endpoint(
    websocket: WebSocket,
    client_id: int | str,
):
    await manager.connect(websocket, client_id, register_client=False)
    await manager.broadcast(f"Client {client_id} joined")
    try:
        while True:
            data = await websocket.receive_text()
            try:
                json_data: dict = json.loads(data)
                if json_data.get("type") == "log":  # and isinstance(json_data, dict)
                    json_data.pop("type")
                    await forward_log(json_data, websocket)
                elif json_data.get("message") == "refresh":
                    await manager.broadcast(json.dumps(json_data))

            except Exception:
                await manager.broadcast(f"Client #{client_id} sent message {data}")
    except WebSocketDisconnect:
        log.info(f"Disconnecting Client {sanitise(str(client_id))}")
        manager.disconnect(client_id, unregister_client=False)
        await manager.broadcast(f"Client #{client_id} disconnected")


async def forward_log(logrecord: dict[str, Any], websocket: WebSocket):
    record_name = logrecord["name"]
    logrecord.pop("msecs", None)
    logrecord.pop("relativeCreated", None)
    client_timestamp = logrecord.pop("created", 0)
    if client_timestamp:
        logrecord["client_time"] = datetime.fromtimestamp(client_timestamp).isoformat()
    logrecord["client_host"] = websocket.client.host
    logging.getLogger(record_name).handle(logging.makeLogRecord(logrecord))


@ws.delete("/connect/{client_id}")
async def close_websocket_connection(client_id: int | str):
    client_id_str = str(client_id).replace("\r\n", "").replace("\n", "")
    log.info(f"Disconnecting {client_id_str}")
    manager.disconnect(client_id)
    await manager.broadcast(f"Client #{client_id} disconnected")



import logging
import sys
from typing import Optional
from unittest.mock import MagicMock
from urllib.parse import urlparse

import pytest
import uvicorn
from fastapi import FastAPI
from fastapi.testclient import TestClient
from packaging.version import Version
from pytest_mock import MockerFixture

import murfey
from murfey.client.update import UPDATE_SUCCESS
from murfey.instrument_server import check_for_updates, start_instrument_server
from murfey.server.api.bootstrap import pypi as pypi_router, version as version_router
from murfey.util.api import url_path_for
from murfey.util.logging import HTTPSHandler

# Set up a test router with only the essential endpoints
app = FastAPI()
for router in [pypi_router, version_router]:
    app.include_router(router)
client = TestClient(app)
base_url = str(client.base_url)


check_for_updates_test_matrix = (
    # Downgrade, upgrade, or keep client version?
    ("downgrade",),
    ("upgrade",),
    ("keep",),
)


@pytest.mark.parametrize("test_params", check_for_updates_test_matrix)
def test_check_for_updates(
    test_params: tuple[str],
    mocker: MockerFixture,
):
    # Unpack test params
    (bump_client_version,) = test_params

    # Modify client version as needed
    current_version = murfey.__version__
    supported_client_version = murfey.__supported_client_version__

    major, minor, patch = Version(current_version).release

    # Adjust the perceived client version in the function being tested
    if bump_client_version == "downgrade":
        support_client_version_parts = Version(supported_client_version).release
        if patch == 0:
            if minor == 0:
                if major == 0:
                    # This can't be downgraded, so skip
                    pytest.skip("This version can't be downgraded anymore; skipping")
                else:
                    major = support_client_version_parts[0] - 1
                    print(f"Downgraded major version to {major}")
            else:
                minor = support_client_version_parts[1] - 1
                print(f"Downgraded minor version to {minor}")
        else:
            patch = support_client_version_parts[2] - 1
            print(f"Downgraded patch version to {patch}")
    elif bump_client_version == "upgrade":
        patch += 1
        print(f"Bumped patch version to {patch}")
    mock_client_version = f"{major}.{minor}.{patch}"

    # Run the version check query and get a response to patch in later
    api_base = urlparse(base_url, allow_fragments=False)
    proxy_path = api_base.path.rstrip("/")
    version_check_path = url_path_for("bootstrap.version", "get_version")
    version_check_query = f"client_version={mock_client_version}"
    version_check_url = api_base._replace(
        path=f"{proxy_path}{version_check_path}",
        query=version_check_query,
    )
    version_check_response = client.get(f"{version_check_path}?{version_check_query}")

    # Check that the endpoint works as expected
    assert version_check_response.status_code == 200
    assert version_check_response.json() == {
        "server": current_version,
        "oldest-supported-client": supported_client_version,
        "client-needs-update": True if bump_client_version == "downgrade" else False,
        "client-needs-downgrade": True if bump_client_version == "upgrade" else False,
    }

    # Patch the URL parse result
    mock_parse = mocker.patch("murfey.instrument_server.urlparse")
    mock_parse.return_value = api_base

    # Patch the result of get
    mock_get = mocker.patch("murfey.client.update.requests.get")
    mock_get.return_value = version_check_response

    # Patch the installation function
    mock_install = mocker.patch("murfey.client.update.install_murfey")

    # Patch the perceived client version
    mocker.patch("murfey.client.update.murfey.__version__", new=mock_client_version)

    # If changing the client version, check that 'install_murfey' and 'exit' are called
    if bump_client_version in ("upgrade", "downgrade"):
        with pytest.raises(SystemExit) as exc_info:
            check_for_updates()
        mock_install.assert_called_once()
        # Check that 'exit' is called with the correct message
        assert exc_info.value.code == UPDATE_SUCCESS
    # If client version is the same, 'install_murfey' shouldn't be called
    else:
        check_for_updates()
        mock_install.assert_not_called()

    # Check that the query URL is correct
    mock_get.assert_called_once_with(version_check_url.geturl())


start_instrument_server_test_matrix = (
    # Host | Port
    (
        None,
        None,
    ),  # Test default values
    (
        "127.0.0.1",
        8000,
    ),  # Test manually included values
)


@pytest.mark.parametrize("test_params", start_instrument_server_test_matrix)
def test_start_instrument_server(
    mocker: MockerFixture,
    mock_client_configuration,
    test_params: tuple[Optional[str], Optional[int]],
):
    # Unpack test params
    host, port = test_params

    # Patch the 'read_config' function
    _ = mocker.patch(
        "murfey.util.client.read_config", return_value=mock_client_configuration
    )

    # Mock the HTTPSHandler (test it separately in a unit test)
    mock_https_handler_instance = MagicMock()
    mock_https_handler_instance.level = logging.INFO
    mock_https_handler_instance.setLevel.return_value = None
    mock_https_handler = mocker.patch(
        "murfey.util.logging.HTTPSHandler",
        spec=HTTPSHandler,
    )
    mock_https_handler.return_value = mock_https_handler_instance

    # Patch the Uvicorn Server instance
    mock_server = mocker.patch("uvicorn.Server")
    # Disable 'run'; we just want to confirm it's called correctly
    mock_server.run.return_value = lambda: None

    # Construct the expected Uvicorn Config object and save it as a dict
    expected_config = vars(
        uvicorn.Config(
            "murfey.instrument_server.main:app",
            host=host if host is not None else "0.0.0.0",
            port=port if port is not None else 8001,
            log_config=None,
            ws_ping_interval=300,
            ws_ping_timeout=300,
        )
    )

    # Construct the arguments to pass to the instrument server
    sys.argv = [
        "murfey.instrument_server",
    ]

    # Add host and port if they're present
    if host is not None:
        sys.argv.extend(["--host", host])
    if port is not None:
        sys.argv.extend(["--port", str(port)])

    # Run the function
    start_instrument_server()

    # Check that the server was called with the correct arguments
    args, kwargs = mock_server.call_args
    actual_config = vars(kwargs["config"])
    assert expected_config == actual_config



from __future__ import annotations

import logging
from itertools import count
from pathlib import Path
from typing import Any, Optional, OrderedDict

import xmltodict

from murfey.client.context import Context, ProcessingParameter
from murfey.client.destinations import find_longest_data_directory
from murfey.client.instance_environment import (
    MovieTracker,
    MurfeyID,
    MurfeyInstanceEnvironment,
)
from murfey.util.client import capture_get, capture_post, get_machine_config_client
from murfey.util.spa_metadata import (
    foil_hole_data,
    foil_hole_from_file,
    get_grid_square_atlas_positions,
    grid_square_data,
    grid_square_from_file,
)

logger = logging.getLogger("murfey.client.contexts.spa")


def _file_transferred_to(
    environment: MurfeyInstanceEnvironment, source: Path, file_path: Path, token: str
):
    machine_config = get_machine_config_client(
        str(environment.url.geturl()),
        token,
        instrument_name=environment.instrument_name,
    )
    if environment.visit in environment.default_destinations[source]:
        return (
            Path(machine_config.get("rsync_basepath", ""))
            / Path(environment.default_destinations[source])
            / file_path.relative_to(source)  # need to strip out the rsync_module name
        )
    return (
        Path(machine_config.get("rsync_basepath", ""))
        / Path(environment.default_destinations[source])
        / environment.visit
        / file_path.relative_to(source)
    )


def _grid_square_metadata_file(
    f: Path, data_directories: list[Path], visit: str, grid_square: int
) -> Path:
    base_dir, mid_dir = find_longest_data_directory(f, data_directories)
    if not base_dir or not mid_dir:
        raise ValueError(f"Could not determine grid square metadata path for {f}")
    metadata_file = (
        base_dir
        / visit
        / mid_dir.parent.parent.parent
        / "Metadata"
        / f"GridSquare_{grid_square}.dm"
    )
    if not metadata_file.is_file():
        logger.warning(f"Grid square metadata file {str(metadata_file)} does not exist")
    return metadata_file


def _get_source(file_path: Path, environment: MurfeyInstanceEnvironment) -> Path | None:
    possible_sources = []
    for s in environment.sources:
        if file_path.is_relative_to(s):
            possible_sources.append(s)
    if not possible_sources:
        return None
    elif len(possible_sources) == 1:
        return possible_sources[0]
    source = possible_sources[0]
    for extra_source in possible_sources[1:]:
        if extra_source.is_relative_to(source):
            source = extra_source
    return source


def _get_xml_list_index(key: str, xml_list: list) -> int:
    for i, elem in enumerate(xml_list):
        if elem["a:Key"] == key:
            return i
    raise ValueError(f"Key not found in XML list: {key}")


class SPAModularContext(Context):
    user_params = [
        ProcessingParameter(
            "dose_per_frame",
            "Dose Per Frame [e- / Angstrom^2 / frame] (after EER grouping if relevant)",
            default=1,
        ),
        ProcessingParameter("symmetry", "Symmetry Group", default="C1"),
        ProcessingParameter("eer_fractionation", "EER Fractionation", default=20),
        ProcessingParameter("gain_ref", "Gain Reference"),
    ]
    metadata_params = [
        ProcessingParameter("voltage", "Voltage"),
        ProcessingParameter("image_size_x", "Image Size X"),
        ProcessingParameter("image_size_y", "Image Size Y"),
        ProcessingParameter("pixel_size_on_image", "Pixel Size"),
        ProcessingParameter("motion_corr_binning", "Motion Correction Binning"),
    ]

    def __init__(self, acquisition_software: str, basepath: Path, token: str):
        super().__init__("SPA", acquisition_software, token)
        self._basepath = basepath
        self._processing_job_stash: dict = {}
        self._foil_holes: dict[int, list[int]] = {}

    def gather_metadata(
        self, metadata_file: Path, environment: MurfeyInstanceEnvironment | None = None
    ):
        logger.info(f"trying to gather metadata on {metadata_file}")
        if metadata_file.suffix != ".xml":
            raise ValueError(
                f"SPA gather_metadata method expected xml file not {metadata_file.name}"
            )
        if not metadata_file.is_file():
            logger.debug(f"Metadata file {metadata_file} not found")
            return OrderedDict({})
        with open(metadata_file, "r") as xml:
            try:
                for_parsing = xml.read()
            except Exception:
                logger.warning(f"Failed to parse file {metadata_file}")
                return OrderedDict({})
            data = xmltodict.parse(for_parsing)
        magnification = 0
        metadata: OrderedDict = OrderedDict({})
        metadata["experiment_type"] = "SPA"
        if data.get("Acquisition"):
            metadata["voltage"] = 300
            metadata["image_size_x"] = data["Acquisition"]["Info"]["ImageSize"]["Width"]
            metadata["image_size_y"] = data["Acquisition"]["Info"]["ImageSize"][
                "Height"
            ]
            metadata["pixel_size_on_image"] = float(
                data["Acquisition"]["Info"]["SensorPixelSize"]["Height"]
            )
            metadata["magnification"] = magnification
        elif data.get("MicroscopeImage"):
            metadata["voltage"] = (
                float(
                    data["MicroscopeImage"]["microscopeData"]["gun"][
                        "AccelerationVoltage"
                    ]
                )
                / 1000
            )
            metadata["image_size_x"] = data["MicroscopeImage"]["microscopeData"][
                "acquisition"
            ]["camera"]["ReadoutArea"]["a:width"]
            metadata["image_size_y"] = data["MicroscopeImage"]["microscopeData"][
                "acquisition"
            ]["camera"]["ReadoutArea"]["a:height"]
            metadata["pixel_size_on_image"] = float(
                data["MicroscopeImage"]["SpatialScale"]["pixelSize"]["x"][
                    "numericValue"
                ]
            )
            magnification = data["MicroscopeImage"]["microscopeData"]["optics"][
                "TemMagnification"
            ]["NominalMagnification"]
            metadata["magnification"] = magnification
            try:
                dose_index = _get_xml_list_index(
                    "Dose",
                    data["MicroscopeImage"]["CustomData"]["a:KeyValueOfstringanyType"],
                )
                metadata["total_exposed_dose"] = round(
                    float(
                        data["MicroscopeImage"]["CustomData"][
                            "a:KeyValueOfstringanyType"
                        ][dose_index]["a:Value"]["#text"]
                    )
                    * (1e-20),
                    2,
                )  # convert e / m^2 to e / A^2
            except ValueError:
                metadata["total_exposed_dose"] = 1
            c2_index = 3
            for i, el in enumerate(
                data["MicroscopeImage"]["CustomData"]["a:KeyValueOfstringanyType"]
            ):
                if el["a:Key"] == "Aperture[C2].Name":
                    c2_index = i
                    break
            metadata["c2aperture"] = data["MicroscopeImage"]["CustomData"][
                "a:KeyValueOfstringanyType"
            ][c2_index]["a:Value"]["#text"]
            metadata["exposure_time"] = data["MicroscopeImage"]["microscopeData"][
                "acquisition"
            ]["camera"]["ExposureTime"]
            try:
                metadata["slit_width"] = data["MicroscopeImage"]["microscopeData"][
                    "optics"
                ]["EnergyFilter"]["EnergySelectionSlitWidth"]
            except KeyError:
                metadata["slit_width"] = None
            metadata["phase_plate"] = (
                1
                if data["MicroscopeImage"]["CustomData"]["a:KeyValueOfstringanyType"][
                    11
                ]["a:Value"]["#text"]
                == "true"
                else 0
            )
        else:
            logger.warning("Metadata file format is not recognised")
            return OrderedDict({})
        binning_factor_xml = int(
            data["MicroscopeImage"]["microscopeData"]["acquisition"]["camera"][
                "Binning"
            ]["a:x"]
        )
        binning_factor = 1
        if environment:
            server_config_response = capture_get(
                base_url=str(environment.url.geturl()),
                router_name="session_control.router",
                function_name="machine_info_by_instrument",
                token=self._token,
                instrument_name=environment.instrument_name,
            )
            if server_config_response is None:
                return None
            server_config = server_config_response.json()
            if server_config.get("superres") and not environment.superres:
                # If camera is capable of superres and collection is in superres
                binning_factor = 2
            elif not server_config.get("superres"):
                binning_factor_xml = 2
            if magnification:
                ps_from_mag = (
                    server_config.get("calibrations", {})
                    .get("magnification", {})
                    .get(magnification)
                )
                if ps_from_mag:
                    metadata["pixel_size_on_image"] = float(ps_from_mag) * 1e-10
                    # this is a bit of a hack to cover the case when the data is binned K3
                    # then the pixel size from the magnification table will be correct but the binning_factor will be 2
                    # this is divided out later so multiply it in here to cancel
                    if server_config.get("superres") and not environment.superres:
                        metadata["pixel_size_on_image"] /= (
                            1 if binning_factor_xml == 2 else 2
                        )
                else:
                    metadata["pixel_size_on_image"] /= binning_factor
        metadata["image_size_x"] = str(int(metadata["image_size_x"]) * binning_factor)
        metadata["image_size_y"] = str(int(metadata["image_size_y"]) * binning_factor)
        metadata["motion_corr_binning"] = 1 if binning_factor_xml == 2 else 2
        metadata["gain_ref"] = environment.gain_ref if environment else None
        metadata["dose_per_frame"] = environment.dose_per_frame if environment else None
        metadata["symmetry"] = (environment.symmetry if environment else None) or "C1"
        metadata["eer_fractionation"] = (
            environment.eer_fractionation if environment else None
        ) or 20
        metadata["source"] = str(self._basepath)
        return metadata

    def _position_analysis(
        self,
        transferred_file: Path,
        environment: MurfeyInstanceEnvironment,
        source: Path,
        machine_config: dict,
    ) -> Optional[int]:
        grid_square = grid_square_from_file(transferred_file)
        grid_square_metadata_file = _grid_square_metadata_file(
            transferred_file,
            [Path(p) for p in machine_config["data_directories"]],
            environment.visit,
            grid_square,
        )
        if (
            grid_square is not None
            and environment.murfey_session is not None
            and self._foil_holes.get(grid_square) is None
        ):
            self._foil_holes[grid_square] = []
            gs_pix_position: tuple[
                Optional[int],
                Optional[int],
                Optional[float],
                Optional[float],
                Optional[int],
                Optional[int],
                Optional[float],
            ] = (None, None, None, None, None, None, None)
            data_collection_group = (
                capture_get(
                    base_url=str(environment.url.geturl()),
                    router_name="session_control.router",
                    function_name="get_dc_groups",
                    token=self._token,
                    session_id=environment.murfey_session,
                )
                .json()
                .get(str(source), {})
            )
            if not data_collection_group:
                logger.info("Data collection group has not yet been made")
                return None
            if data_collection_group.get("atlas"):
                visit_path = ""
                for p in transferred_file.parts:
                    if p == environment.visit:
                        break
                    visit_path += f"/{p}"
                if source in list(environment.samples.keys()):
                    local_atlas_path = (
                        Path(visit_path) / environment.samples[source].atlas
                    )
                    gs_pix_position = get_grid_square_atlas_positions(
                        local_atlas_path,
                        grid_square=str(grid_square),
                    )[str(grid_square)]
            gs = grid_square_data(
                grid_square_metadata_file,
                grid_square,
            )
            metadata_source_as_str = (
                "/".join(source.parts[:-2])
                + f"/{environment.visit}/"
                + source.parts[-2]
            )
            metadata_source = Path(
                metadata_source_as_str[1:]
                if metadata_source_as_str.startswith("//")
                else metadata_source_as_str
            )
            image_path = (
                _file_transferred_to(
                    environment, metadata_source, Path(gs.image), self._token
                )
                if gs.image
                else ""
            )
            capture_post(
                base_url=str(environment.url.geturl()),
                router_name="session_control.spa_router",
                function_name="register_grid_square",
                token=self._token,
                session_id=environment.murfey_session,
                gsid=grid_square,
                data={
                    "tag": str(source),
                    "readout_area_x": gs.readout_area_x,
                    "readout_area_y": gs.readout_area_y,
                    "thumbnail_size_x": gs.thumbnail_size_x,
                    "thumbnail_size_y": gs.thumbnail_size_y,
                    "pixel_size": gs.pixel_size,
                    "image": str(image_path),
                    "x_location": gs_pix_position[0],
                    "y_location": gs_pix_position[1],
                    "x_stage_position": gs_pix_position[2],
                    "y_stage_position": gs_pix_position[3],
                    "width": gs_pix_position[4],
                    "height": gs_pix_position[5],
                    "angle": gs_pix_position[6],
                },
            )
        foil_hole = foil_hole_from_file(transferred_file)
        if foil_hole not in self._foil_holes[grid_square]:
            if environment.murfey_session is not None:
                fh = foil_hole_data(
                    grid_square_metadata_file,
                    foil_hole,
                    grid_square,
                )
                metadata_source_as_str = (
                    "/".join(source.parts[:-2])
                    + f"/{environment.visit}/"
                    + source.parts[-2]
                )
                metadata_source = Path(
                    metadata_source_as_str[1:]
                    if metadata_source_as_str.startswith("//")
                    else metadata_source_as_str
                )
                image_path = (
                    _file_transferred_to(
                        environment, metadata_source, Path(fh.image), self._token
                    )
                    if fh.image
                    else ""
                )
                capture_post(
                    base_url=str(environment.url.geturl()),
                    router_name="session_control.spa_router",
                    function_name="register_foil_hole",
                    token=self._token,
                    session_id=environment.murfey_session,
                    gs_name=grid_square,
                    data={
                        "name": foil_hole,
                        "x_location": fh.x_location,
                        "y_location": fh.y_location,
                        "x_stage_position": fh.x_stage_position,
                        "y_stage_position": fh.y_stage_position,
                        "readout_area_x": fh.readout_area_x,
                        "readout_area_y": fh.readout_area_y,
                        "thumbnail_size_x": fh.thumbnail_size_x,
                        "thumbnail_size_y": fh.thumbnail_size_y,
                        "pixel_size": fh.pixel_size,
                        "diameter": fh.diameter,
                        "tag": str(source),
                        "image": str(image_path),
                    },
                )
            else:
                capture_post(
                    base_url=str(environment.url.geturl()),
                    router_name="session_control.spa_router",
                    function_name="register_foil_hole",
                    token=self._token,
                    session_id=environment.murfey_session,
                    gs_name=grid_square,
                    data={
                        "name": foil_hole,
                        "tag": str(source),
                    },
                )
            self._foil_holes[grid_square].append(foil_hole)
        return foil_hole

    def post_transfer(
        self,
        transferred_file: Path,
        environment: MurfeyInstanceEnvironment | None = None,
        **kwargs,
    ) -> bool:
        super().post_transfer(
            transferred_file=transferred_file,
            environment=environment,
            **kwargs,
        )
        data_suffixes = (".mrc", ".tiff", ".tif", ".eer")
        if "gain" not in transferred_file.name:
            if transferred_file.suffix in data_suffixes:
                if self._acquisition_software == "epu":
                    if environment:
                        machine_config = get_machine_config_client(
                            str(environment.url.geturl()),
                            self._token,
                            instrument_name=environment.instrument_name,
                        )
                    else:
                        machine_config = {}
                    required_strings = (
                        machine_config.get("data_required_substrings", {})
                        .get("epu", {})
                        .get(transferred_file.suffix, ["fractions"])
                    )

                    if not environment:
                        logger.warning("No environment passed in")
                        return True
                    source = _get_source(transferred_file, environment)
                    if not source:
                        logger.warning(f"No source found for file {transferred_file}")
                        return True

                    if required_strings and not any(
                        r in transferred_file.name for r in required_strings
                    ):
                        return True

                    if environment:
                        file_transferred_to = _file_transferred_to(
                            environment, source, transferred_file, self._token
                        )
                        if not environment.movie_counters.get(str(source)):
                            movie_counts_get = capture_get(
                                base_url=str(environment.url.geturl()),
                                router_name="session_control.router",
                                function_name="count_number_of_movies",
                                token=self._token,
                            )
                            if movie_counts_get is not None:
                                environment.movie_counters[str(source)] = count(
                                    movie_counts_get.json().get(str(source), 0) + 1
                                )
                        environment.movies[file_transferred_to] = MovieTracker(
                            movie_number=next(environment.movie_counters[str(source)]),
                            motion_correction_uuid=next(MurfeyID),
                        )

                        eer_fractionation_file = None
                        if file_transferred_to.suffix == ".eer":
                            response = capture_post(
                                base_url=str(environment.url.geturl()),
                                router_name="file_io_instrument.router",
                                function_name="write_eer_fractionation_file",
                                token=self._token,
                                visit_name=environment.visit,
                                session_id=environment.murfey_session,
                                data={
                                    "eer_path": str(file_transferred_to),
                                    "fractionation": self.data_collection_parameters[
                                        "eer_fractionation"
                                    ],
                                    "dose_per_frame": self.data_collection_parameters[
                                        "dose_per_frame"
                                    ],
                                    "fractionation_file_name": "eer_fractionation_spa.txt",
                                },
                            )
                            if response is None:
                                return False
                            eer_fractionation_file = response.json()[
                                "eer_fractionation_file"
                            ]

                        try:
                            foil_hole: Optional[int] = self._position_analysis(
                                transferred_file, environment, source, machine_config
                            )
                        except Exception as e:
                            # try to continue if position information gathering fails so that movie is processed anyway
                            logger.warning(
                                f"Unable to register foil hole for {str(file_transferred_to)}. Exception: {str(e)}",
                                exc_info=True,
                            )
                            foil_hole = None

                        preproc_data = {
                            "path": str(file_transferred_to),
                            "description": "",
                            "processing_job": None,
                            "data_collection_id": None,
                            "image_number": environment.movies[
                                file_transferred_to
                            ].movie_number,
                            "pixel_size": self.data_collection_parameters.get(
                                "pixel_size_on_image"
                            ),
                            "autoproc_program_id": None,
                            "dose_per_frame": environment.dose_per_frame,
                            "mc_binning": self.data_collection_parameters.get(
                                "motion_corr_binning", 1
                            ),
                            "gain_ref": environment.gain_ref,
                            "extract_downscale": self.data_collection_parameters.get(
                                "downscale", True
                            ),
                            "eer_fractionation_file": eer_fractionation_file,
                            "tag": str(source),
                            "foil_hole_id": foil_hole,
                        }
                        capture_post(
                            base_url=str(environment.url.geturl()),
                            router_name="workflow.spa_router",
                            function_name="request_spa_preprocessing",
                            token=self._token,
                            visit_name=environment.visit,
                            session_id=environment.murfey_session,
                            data={
                                k: None if v == "None" else v
                                for k, v in preproc_data.items()
                            },
                        )

        return True

    def _register_data_collection(
        self,
        tag: str,
        url: str,
        data: dict,
        environment: MurfeyInstanceEnvironment,
    ):
        return

    def _register_processing_job(
        self,
        tag: str,
        environment: MurfeyInstanceEnvironment,
        parameters: dict[str, Any] | None = None,
    ):
        return

    def _launch_spa_pipeline(
        self,
        tag: str,
        jobid: int,
        environment: MurfeyInstanceEnvironment,
        url: str = "",
    ):
        return



"""
rsync is a Linux process that does fast, versatile remote (and local) file copying.
This module allows Murfey to use rsync to facilitate its file transfers, and to
parse the logs outputted by rsync in order confirm that the file has been correctly
transferred.
"""

from __future__ import annotations

import logging
import os
import queue
import shutil
import subprocess
import threading
import time
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Awaitable, Callable, List, NamedTuple
from urllib.parse import ParseResult

from murfey.util.client import Observer

logger = logging.getLogger("murfey.client.rsync")

# $ rsync -riiv . --times --outbuf=line wra62962@ws133:/dls/tmp/wra62962/junk --dry-run
# sending incremental file list
# .d          ./
# .f          README.md
# <f.st...... large
# .f          tests/util/__pycache__/test_state.cpython-39-pytest-6.2.5.pyc
# <f+++++++++ tests/server/test_main.py
#
# sent 3,785 bytes  received 355 bytes  2,760.00 bytes/sec
# total size is 314,923,092  speedup is 76,068.38 (DRY RUN)


class TransferResult(Enum):
    SUCCESS = 1
    FAILURE = 2


class RSyncerUpdate(NamedTuple):
    file_path: Path
    file_size: int
    outcome: TransferResult
    transfer_total: int
    queue_size: int
    base_path: Path | None = None


class RSyncer(Observer):
    def __init__(
        self,
        basepath_local: Path,
        basepath_remote: Path,
        rsync_module: str,
        server_url: ParseResult,
        stop_callback: Callable = lambda *args, **kwargs: None,
        local: bool = False,
        do_transfer: bool = True,
        remove_files: bool = False,
        required_substrings_for_removal: List[str] = [],
        substrings_blacklist: dict[str, list[str]] = {},
        notify: bool = True,
        end_time: datetime | None = None,
    ):
        super().__init__()
        self._basepath = basepath_local.absolute()
        self._basepath_remote = basepath_remote
        self._rsync_module = rsync_module
        self._server_url = server_url
        self._stop_callback = stop_callback
        self._local = local
        self._do_transfer = do_transfer
        self._remove_files = remove_files
        self._required_substrings_for_removal = required_substrings_for_removal
        self._substrings_blacklist = substrings_blacklist
        self._notify = notify
        self._end_time = end_time

        self._skipped_files: List[Path] = []

        # Set rsync destination
        self._remote = (
            str(basepath_remote)
            if local
            else f"{server_url.hostname}::{self._rsync_module}/{basepath_remote}/"
        )
        logger.debug(f"rsync destination path set to {self._remote}")

        # For local tests you can use something along the lines of
        # self._remote = f"wra62962@ws133:/dls/tmp/wra62962/junk/{basepath_remote}"
        # to avoid having to set up an rsync daemon
        self._files_transferred = 0
        self._bytes_transferred = 0

        # self.queue = queue.Queue[Optional[Path]]()
        self.queue: queue.Queue[Path | None] = queue.Queue()
        self.thread = threading.Thread(
            name=f"RSync {self._basepath}:{self._remote}",
            target=self._process,
            daemon=True,
        )
        self._stopping = False
        self._halt_thread = False
        self._finalising = False
        self._finalised = False

    @property
    def status(self) -> str:
        if self._stopping:
            if self.thread.is_alive():
                return "stopping"
            else:
                return "finished"
        else:
            if self.thread.is_alive():
                return "running"
            else:
                return "ready"

    def __repr__(self) -> str:
        return f"<RSyncer ({self._basepath} → {self._remote})>"

    @classmethod
    def from_rsyncer(cls, rsyncer: RSyncer, **kwargs):
        kwarguments_from_rsyncer = {
            "local": rsyncer._local,
            "do_transfer": rsyncer._do_transfer,
            "remove_files": rsyncer._remove_files,
            "notify": rsyncer._notify,
        }
        kwarguments_from_rsyncer.update(kwargs)
        return cls(
            rsyncer._basepath,
            rsyncer._basepath_remote,
            rsyncer._rsync_module,
            rsyncer._server_url,
            local=kwarguments_from_rsyncer["local"],
            do_transfer=kwarguments_from_rsyncer["do_transfer"],
            remove_files=kwarguments_from_rsyncer["remove_files"],
            notify=kwarguments_from_rsyncer["notify"],
        )

    def notify(self, *args, secondary: bool = False, **kwargs) -> None:
        if self._notify:
            super().notify(*args, secondary=secondary, **kwargs)

    def start(self):
        if self.thread.is_alive():
            raise RuntimeError("RSyncer already running")
        if self._stopping:
            raise RuntimeError("RSyncer has already stopped")
        logger.info(f"RSync thread starting for {self}")
        self.thread.start()

    def restart(self):
        self.thread.join()
        self._halt_thread = False
        self.thread = threading.Thread(
            name=f"RSync {self._basepath}:{self._remote}",
            target=self._process,
            daemon=True,
        )
        self.start()

    def stop(self):
        logger.info(f"Stopping RSync thread {self}")
        self._stopping = True
        if self.thread.is_alive():
            logger.info("Waiting for ongoing transfers to complete...")
            self.queue.join()

        self._halt_thread = True
        if self.thread.is_alive():
            self.queue.put(None)
            self.thread.join()
        logger.info(f"RSync thread {self} successfully stopped")

    def request_stop(self):
        self._stopping = True
        self._halt_thread = True

    def finalise(
        self,
        thread: bool = True,
        callback: Callable[..., Awaitable[None] | None] | None = None,
    ):
        self.stop()
        self._remove_files = True
        self._notify = False
        self._end_time = None
        self._finalising = True

        # Perform recursive cleanup on current directory
        logger.info(f"Starting file cleanup for RSync thread {self}")
        files_to_transfer: list[Path] = []

        def recursive_cleanup(dirpath: str | Path):
            for entry in os.scandir(dirpath):
                if entry.is_dir():
                    # Recursively delete directories with blacklisted substrings
                    if any(
                        pattern in entry.name
                        for pattern in self._substrings_blacklist.get("directories", [])
                    ):
                        logger.debug(f"Deleting blacklisted directory {entry.path}")
                        shutil.rmtree(entry.path)
                        continue
                    # Recursively search in whitelisted ones
                    recursive_cleanup(entry.path)
                elif entry.is_file():
                    # Delete blacklisted files
                    if any(
                        pattern in entry.name
                        for pattern in self._substrings_blacklist.get("files", [])
                    ):
                        logger.debug(f"Deleting blacklisted file {entry.path}")
                        Path(entry.path).unlink()
                        continue
                    # Append others for transfer
                    files_to_transfer.append(Path(entry.path))

        recursive_cleanup(self._basepath)
        logger.debug(f"Number of files to transfer: {len(files_to_transfer)}")

        if thread:
            self.thread = threading.Thread(
                name=f"RSync finalisation {self._basepath}:{self._remote}",
                target=self._process,
                daemon=True,
            )
            for f in files_to_transfer:
                self.queue.put(f)
            self.stop()
        else:
            self._transfer(files_to_transfer)
        self._finalised = True
        logger.info(f"File cleanup for RSync thread {self} successfully completed")
        if callback:
            callback()

    def enqueue(self, file_path: Path):
        if not self._stopping:
            absolute_path = self._basepath / file_path
            self.queue.put(absolute_path)

    def flush_skipped(self):
        for f in self._skipped_files:
            self.queue.put(f)
        self._skipped_files = []

    def _process(self):
        logger.info(f"Starting main process loop for RSync thread {self}")
        files_to_transfer: list[Path]
        backoff = 0
        while not self._halt_thread:
            first = self.queue.get()
            if not first:
                # allow leaving thread when 'None' is passed
                self.queue.task_done()
                continue

            files_to_transfer = [first] if not first.name.startswith(".") else []
            stop = False
            try:
                num_files = 0
                while True:
                    if num_files > 100:
                        break
                    next_file = self.queue.get(block=True, timeout=0.1)
                    if not next_file:
                        stop = True
                        break
                    if not next_file.name.startswith("."):
                        files_to_transfer.append(next_file)
                        num_files += 1
            except queue.Empty:
                pass

            logger.info(f"Preparing to transfer {len(files_to_transfer)} files")
            if self._do_transfer:
                try:
                    success = self._transfer(files_to_transfer)
                except Exception as e:
                    logger.error(
                        f"Unhandled exception {e} in RSync thread", exc_info=True
                    )
                    success = False
            else:
                success = self._fake_transfer(files_to_transfer)

            logger.info(f"Completed transfer of {len(files_to_transfer)} files")
            for _ in files_to_transfer:
                self.queue.task_done()
            logger.debug(
                f"{self.queue.unfinished_tasks} files remain in queue for processing"
            )

            if success:
                backoff = 0
            else:
                backoff = min(backoff * 2 + 1, 120)
                logger.info(f"Waiting {backoff} seconds before next rsync attempt")
                time.sleep(backoff)

            if stop:
                self.queue.task_done()
                continue

        self._stop_callback(self._basepath, explicit_stop=self._stopping)

    def _fake_transfer(self, files: list[Path]) -> bool:
        previously_transferred = self._files_transferred

        relative_filenames = []
        for f in files:
            try:
                relative_filenames.append(f.relative_to(self._basepath))
            except ValueError:
                raise ValueError(f"File '{f}' is outside of {self._basepath}") from None

        updates = []
        for f in set(relative_filenames):
            self._files_transferred += 1
            update = RSyncerUpdate(
                file_path=f,
                file_size=0,
                outcome=TransferResult.SUCCESS,
                transfer_total=self._files_transferred - previously_transferred,
                queue_size=0,
                base_path=self._basepath,
            )
            self.notify(update)
            updates.append(update)
            time.sleep(0.01)
            self.notify([update], num_skipped_files=0, secondary=True)
        # self.notify(updates, secondary=True)

        return True

    def _transfer(self, infiles: list[Path]) -> bool:
        """
        Transfer files via an rsync sub-process, and parses the rsync stdout to verify
        the success of the transfer.
        """

        # Set up initial variables
        if self._end_time:
            files = [
                f
                for f in infiles
                if f.is_file() and f.stat().st_ctime < self._end_time.timestamp()
            ]
            self._skipped_files.extend(set(infiles).difference(set(files)))
            num_skipped_files = len(set(infiles).difference(set(files)))
        elif self._finalising:
            files = [f for f in infiles if f.is_file() and f not in self._skipped_files]
            num_skipped_files = 0
        else:
            files = [f for f in infiles if f.is_file()]
            num_skipped_files = 0

        previously_transferred = self._files_transferred
        transfer_success: set[Path] = set()
        successful_updates: list[RSyncerUpdate] = []
        next_file: RSyncerUpdate | None = None

        def parse_stdout(line: str):
            """
            Reads the stdout from rsync in order to verify the status of transferred
            files and other things.
            """
            nonlocal next_file

            # Miscellaneous rsync stdout lines to skip
            if not line:
                return
            if line.startswith(("building file list", "created directory", "sending")):
                return
            if line.startswith("sent "):
                # sent 6,676 bytes  received 397 bytes  4,715.33 bytes/sec
                return
            if line.startswith("total "):
                # total size is 315,265,653  speedup is 44,573.12 (DRY RUN)
                return
            if line.startswith(("cd", ".d")):
                return

            # Lines with chr(13), \r, contain file transfer information
            if chr(13) in line:
                # partial transfer
                #
                #           3,136 100%    1.50MB/s    0:00:00
                #           3,136 100%    1.50MB/s    0:00:00 (xfr#5, to-chk=109/115)
                xfer_line = line.split(chr(13))[-1]
                if xfer_line.endswith((" file to consider", " files to consider")):
                    return
                if "(xfr" not in xfer_line:
                    raise RuntimeError(
                        f"Unexpected line {xfer_line} {line.split(chr(13))}"
                    )
                if next_file is None:
                    logger.warning(f"Invalid state {xfer_line=}, {next_file=}")
                    return
                transfer_success.add(next_file.file_path)
                size_bytes = int(xfer_line.split()[0].replace(",", ""))
                self.notify(next_file._replace(file_size=size_bytes))
                successful_updates.append(next_file._replace(file_size=size_bytes))
                next_file = None
                return

            # Lines starting with "*f" contain file transfer information
            if line.startswith((".f", ">f", "<f")):
                # .d          ./
                # .f          README.md
                # .f          tests/util/__pycache__/test_state.cpython-39-pytest-6.2.5.pyc
                # No transfer happening
                if next_file is not None:
                    logger.warning(f"Invalid state {line=}, {next_file=}")
                    return

                self._files_transferred += 1
                current_outstanding = self.queue.unfinished_tasks - (
                    self._files_transferred - previously_transferred
                )
                update = RSyncerUpdate(
                    file_path=Path(
                        line[12:].rstrip()
                    ),  # Remove trailing newlines, spaces, and carriage returns
                    file_size=0,
                    outcome=TransferResult.SUCCESS,
                    transfer_total=self._files_transferred - previously_transferred,
                    queue_size=current_outstanding,
                    base_path=self._basepath,
                )
                if line[0] == ".":
                    # No transfer happening
                    transfer_success.add(update.file_path)
                    self.notify(update)
                    successful_updates.append(update)
                else:
                    # This marks the start of a transfer, wait for the progress line
                    next_file = update
                return
            logger.warning(f"Unknown line from rsync {line}")
            return

        def parse_stderr(line: str):
            if line.strip():
                logger.warning(f"rsync stderr: {line!r}")

        # Generate list of relative filenames for this batch of transferred files
        #   Relative filenames will be safe to use on both Windows and Unix
        relative_filenames: List[Path] = []
        for f in files:
            try:
                relative_filenames.append(f.relative_to(self._basepath))
            except ValueError:
                raise ValueError(f"File '{f}' is outside of {self._basepath}") from None

        # Encode files to rsync as bytestring
        if self._remove_files:
            if self._required_substrings_for_removal:
                rsync_stdin_remove = b"\n".join(
                    os.fsencode(f)
                    for f in relative_filenames
                    if any(
                        substring in f.name
                        for substring in self._required_substrings_for_removal
                    )
                )
                rsync_stdin = b"\n".join(
                    os.fsencode(f)
                    for f in relative_filenames
                    if not any(
                        substring in f.name
                        for substring in self._required_substrings_for_removal
                    )
                )
            else:
                rsync_stdin_remove = b"\n".join(
                    os.fsencode(f) for f in relative_filenames
                )
                rsync_stdin = b""
        else:
            rsync_stdin_remove = b""
            rsync_stdin = b"\n".join(os.fsencode(f) for f in relative_filenames)

        # Create and run rsync subprocesses
        # rsync default settings
        rsync_cmd = [
            "rsync",
            "-iiv",
            "--times",
            "--progress",
            "--outbuf=line",
            "--files-from=-",  # '-' indicates reading from standard input
            # Needed as a pair to trigger permission modifications
            # Ref: https://serverfault.com/a/796341
            "-p",
            "--chmod=D0750,F0750",  # Use extended chmod format
        ]
        # Add file locations
        rsync_cmd.extend([".", self._remote])

        # Transfer files to destination
        result: subprocess.CompletedProcess[bytes] | None = None
        success = True
        if rsync_stdin:
            # Wrap rsync command in a bash command
            cmd = [
                "bash",
                "-c",
                # rsync command passed in as a single string
                " ".join(rsync_cmd),
            ]
            result = subprocess.run(
                cmd,
                cwd=self._basepath,  # As-is Path is fine
                capture_output=True,
                input=rsync_stdin,
            )
            # Parse outputs
            for line in result.stdout.decode("utf-8", "replace").split("\n"):
                parse_stdout(line)
            for line in result.stderr.decode("utf-8", "replace").split("\n"):
                parse_stderr(line)
            success = result.returncode == 0

        # Remove files from source
        if rsync_stdin_remove:
            # Insert file removal flag before locations
            rsync_cmd.insert(-2, "--remove-source-files")
            # Wrap rsync command in a bash command
            cmd = [
                "bash",
                "-c",
                # Pass rsync command as single string
                " ".join(rsync_cmd),
            ]
            result = subprocess.run(
                cmd,
                cwd=self._basepath,
                capture_output=True,
                input=rsync_stdin_remove,
            )
            # Parse outputs
            for line in result.stdout.decode("utf-8", "replace").split("\n"):
                parse_stdout(line)
            for line in result.stderr.decode("utf-8", "replace").split("\n"):
                parse_stderr(line)
            # Leave it as a failure if the previous rsync subprocess failed
            if success:
                success = result.returncode == 0

        self.notify(
            successful_updates, num_skipped_files=num_skipped_files, secondary=True
        )

        # Print out a summary message for each file transfer batch instead of individual messages
        # List out file paths as stored in memory to see if issue is due to file path mismatch
        if len(set(relative_filenames) - transfer_success) != 0:
            logger.debug(
                f"Files identified for transfer ({len(relative_filenames)}): {relative_filenames!r}"
            )
            logger.debug(
                f"Files successfully transferred ({len(transfer_success)}): {list(transfer_success)!r}"
            )

        # Compare files from rsync stdout to original list to verify transfer
        for f in set(relative_filenames) - transfer_success:
            # Mute individual file warnings; replace with summarised one above
            # logger.warning(f"Transfer of file {f.name!r} considered a failure")
            self._files_transferred += 1
            current_outstanding = self.queue.unfinished_tasks - (
                self._files_transferred - previously_transferred
            )
            update = RSyncerUpdate(
                file_path=f,
                file_size=0,
                outcome=TransferResult.FAILURE,
                transfer_total=self._files_transferred,
                queue_size=current_outstanding,
                base_path=self._basepath,
            )
            self.notify(update)
            success = False

        if result is None and files:
            # Only log this as an error if files were scheduled for transfer
            logger.error(f"No rsync process ran for files: {files}")
        elif result and result.returncode:
            logger.warning(
                f"rsync process finished with return code {result.returncode}",
            )
        elif not success:
            logger.info(
                "rsync process failed for some files: "
                f"requested {len(relative_filenames)}, transferred {len(transfer_success)}",
            )
        else:
            logger.debug("rsync process finished successfully")

        return success



from __future__ import annotations

import logging
from pathlib import Path
from threading import RLock
from typing import Callable, Dict, List, OrderedDict

import xmltodict

import murfey.util.eer
from murfey.client.context import Context, ProcessingParameter, ensure_dcg_exists
from murfey.client.instance_environment import (
    MovieID,
    MovieTracker,
    MurfeyID,
    MurfeyInstanceEnvironment,
)
from murfey.util.client import capture_get, capture_post, get_machine_config_client
from murfey.util.mdoc import get_block, get_global_data, get_num_blocks

logger = logging.getLogger("murfey.client.contexts.tomo")


def _get_tilt_angle_v5_7(p: Path) -> str:
    return p.name.split("[")[1].split("]")[0]


def _get_tilt_angle_v5_11(p: Path) -> str:
    _split = p.name.split("_")[2].split(".")
    return ".".join(_split[:-1])


def _find_angle_index(split_name: List[str]) -> int:
    for i, part in enumerate(split_name):
        if "." in part and part[-1].isnumeric():
            return i
    return -1


def _get_tilt_angle_v5_12(p: Path) -> str:
    split_name = p.stem.split("_")
    angle_idx = _find_angle_index(split_name)
    if angle_idx == -1:
        return ""
    return split_name[angle_idx]


tomo_tilt_info = {
    "5.7": _get_tilt_angle_v5_7,
    "5.11": _get_tilt_angle_v5_11,
    "5.12": _get_tilt_angle_v5_12,
}


def _construct_tilt_series_name(file_path: Path) -> str:
    # Assuming files end with _{tiltnumber}_{angle}_{date}_{time}_{fractions}.{suffix}
    split_name = file_path.name.split("_")
    return "_".join(split_name[:-5])


class TomographyContext(Context):
    user_params = [
        ProcessingParameter(
            "dose_per_frame", "Dose Per Frame (e- / Angstrom^2 / frame)", default=1
        ),
        ProcessingParameter("gain_ref", "Gain Reference"),
        ProcessingParameter("eer_fractionation", "EER Fractionation", default=20),
    ]
    metadata_params = [
        ProcessingParameter("voltage", "Voltage"),
        ProcessingParameter("image_size_x", "Image Size X"),
        ProcessingParameter("image_size_y", "Image Size Y"),
        ProcessingParameter("pixel_size_on_image", "Pixel Size"),
        ProcessingParameter("motion_corr_binning", "Motion Correction Binning"),
        ProcessingParameter("frame_count", "Number of image frames"),
        ProcessingParameter("tilt_axis", "Stage rotation angle"),
        ProcessingParameter("num_eer_frames", "Number of EER Frames"),
    ]

    def __init__(self, acquisition_software: str, basepath: Path, token: str):
        super().__init__("Tomography", acquisition_software, token)
        self._basepath = basepath
        self._tilt_series: Dict[str, List[Path]] = {}
        self._tilt_series_with_pjids: List[str] = []
        self._tilt_series_sizes: Dict[str, int] = {}
        self._completed_tilt_series: List[str] = []
        self._aligned_tilt_series: List[str] = []
        self._data_collection_stash: list = []
        self._processing_job_stash: dict = {}
        self._lock: RLock = RLock()

    def register_tomography_data_collections(
        self,
        file_extension: str,
        image_directory: str,
        environment: MurfeyInstanceEnvironment | None = None,
    ):
        if not environment:
            logger.error(
                "No environment passed to register tomography data collections"
            )
            return
        try:
            metadata_source = (
                self._basepath.parent / environment.visit / self._basepath.name
            )
            ensure_dcg_exists(
                collection_type="tomo",
                metadata_source=metadata_source,
                environment=environment,
                token=self._token,
            )

            for tilt_series in self._tilt_series.keys():
                if tilt_series not in self._tilt_series_with_pjids:
                    dc_data = {
                        "experiment_type": "tomography",
                        "file_extension": file_extension,
                        "acquisition_software": self._acquisition_software,
                        "image_directory": image_directory,
                        "data_collection_tag": tilt_series,
                        "source": str(self._basepath),
                        "tag": tilt_series,
                    }
                    if (
                        self.data_collection_parameters
                        and self.data_collection_parameters.get("voltage")
                    ):
                        # Once mdoc parameters are known register processing jobs
                        dc_data.update(
                            {
                                "voltage": self.data_collection_parameters["voltage"],
                                "pixel_size_on_image": self.data_collection_parameters[
                                    "pixel_size_on_image"
                                ],
                                "image_size_x": self.data_collection_parameters[
                                    "image_size_x"
                                ],
                                "image_size_y": self.data_collection_parameters[
                                    "image_size_y"
                                ],
                                "magnification": self.data_collection_parameters[
                                    "magnification"
                                ],
                            }
                        )
                        capture_post(
                            base_url=str(environment.url.geturl()),
                            router_name="workflow.router",
                            function_name="start_dc",
                            token=self._token,
                            visit_name=environment.visit,
                            session_id=environment.murfey_session,
                            data=dc_data,
                        )

                        recipes_to_assign_pjids = [
                            "em-tomo-preprocess",
                            "em-tomo-align",
                        ]
                        if not self._tilt_series_with_pjids:
                            recipes_to_assign_pjids.append("em-tomo-class2d")
                        for recipe in recipes_to_assign_pjids:
                            capture_post(
                                base_url=str(environment.url.geturl()),
                                router_name="workflow.router",
                                function_name="register_proc",
                                token=self._token,
                                visit_name=environment.visit,
                                session_id=environment.murfey_session,
                                data={
                                    "tag": tilt_series,
                                    "source": str(self._basepath),
                                    "recipe": recipe,
                                    "experiment_type": "tomography",
                                },
                            )
                        self._tilt_series_with_pjids.append(tilt_series)
                    else:
                        logger.info(
                            "Cannot register data collection yet as no values from mdoc"
                        )

        except Exception as e:
            logger.error(f"ERROR {e}, {self.data_collection_parameters}", exc_info=True)

    def _file_transferred_to(
        self, environment: MurfeyInstanceEnvironment, source: Path, file_path: Path
    ):
        machine_config = get_machine_config_client(
            str(environment.url.geturl()),
            self._token,
            instrument_name=environment.instrument_name,
        )
        if environment.visit in environment.default_destinations[source]:
            return (
                Path(machine_config.get("rsync_basepath", ""))
                / Path(environment.default_destinations[source])
                / file_path.name
            )
        return (
            Path(machine_config.get("rsync_basepath", ""))
            / Path(environment.default_destinations[source])
            / environment.visit
            / file_path.name
        )

    def _get_source(
        self, file_path: Path, environment: MurfeyInstanceEnvironment
    ) -> Path | None:
        for s in environment.sources:
            if file_path.is_relative_to(s):
                return s
        return None

    def _add_tilt(
        self,
        file_path: Path,
        extract_tilt_angle: Callable[[Path], str],
        environment: MurfeyInstanceEnvironment | None = None,
        required_strings: List[str] | None = None,
    ) -> List[str]:
        if not environment:
            logger.warning("No environment passed in")
            return []
        source = self._get_source(file_path, environment)
        if not source:
            logger.warning(f"No source found for file {file_path}")
            return []
        required_strings = (
            ["fractions"] if required_strings is None else required_strings
        )
        if required_strings and not any(r in file_path.name for r in required_strings):
            return []
        try:
            tilt_angle = extract_tilt_angle(file_path)
            try:
                float(tilt_angle)
            except ValueError:
                logger.error(f"whoops, {tilt_angle}")
                return []
            tilt_series = _construct_tilt_series_name(file_path)

        except Exception:
            logger.debug(
                f"Tilt series and angle could not be determined for {file_path}"
            )
            return []

        if environment:
            file_transferred_to = self._file_transferred_to(
                environment, source, file_path
            )
            environment.movies[file_transferred_to] = MovieTracker(
                movie_number=next(MovieID),
                motion_correction_uuid=next(MurfeyID),
            )
            environment.movie_tilt_pair[file_transferred_to] = tilt_series
            if environment.tilt_angles.get(tilt_series):
                environment.tilt_angles[tilt_series].append(
                    [str(file_transferred_to), tilt_angle]
                )
            else:
                environment.tilt_angles[tilt_series] = [
                    [str(file_transferred_to), tilt_angle]
                ]
        if tilt_series in self._completed_tilt_series:
            logger.warning(
                f"Tilt series {tilt_series} was previously thought complete but now {file_path} has been seen"
            )
            self._completed_tilt_series.remove(tilt_series)
            self._tilt_series_sizes[tilt_series] = 0
            rerun_data = {
                "session_id": environment.murfey_session,
                "tag": tilt_series,
                "source": str(file_path.parent),
            }
            capture_post(
                base_url=str(environment.url.geturl()),
                router_name="workflow.tomo_router",
                function_name="register_tilt_series_for_rerun",
                token=self._token,
                visit_name=environment.visit,
                data=rerun_data,
            )
            if tilt_series in self._aligned_tilt_series:
                with self._lock:
                    self._aligned_tilt_series.remove(tilt_series)

        if not self._tilt_series.get(tilt_series):
            logger.info(f"New tilt series found: {tilt_series}")
            self._tilt_series[tilt_series] = [file_path]
            ts_data = {
                "session_id": environment.murfey_session,
                "tag": tilt_series,
                "source": str(file_path.parent),
            }
            capture_post(
                base_url=str(environment.url.geturl()),
                router_name="workflow.tomo_router",
                function_name="register_tilt_series",
                token=self._token,
                visit_name=environment.visit,
                data=ts_data,
            )
            if not self._tilt_series_sizes.get(tilt_series):
                self._tilt_series_sizes[tilt_series] = 0

            # Will register processing jobs for all tilt series except the first one
            self.register_tomography_data_collections(
                file_extension=file_path.suffix,
                image_directory=str(
                    environment.default_destinations.get(
                        file_path.parent, file_path.parent
                    )
                ),
                environment=environment,
            )
        else:
            if file_path not in self._tilt_series[tilt_series]:
                for p in self._tilt_series[tilt_series]:
                    if tilt_angle == extract_tilt_angle(p):
                        break
                else:
                    self._tilt_series[tilt_series].append(file_path)

        if environment:
            tilt_data = {
                "movie_path": str(file_transferred_to),
                "tilt_series_tag": tilt_series,
                "source": str(file_path.parent),
            }
            capture_post(
                base_url=str(environment.url.geturl()),
                router_name="workflow.tomo_router",
                function_name="register_tilt",
                token=self._token,
                visit_name=environment.visit,
                session_id=environment.murfey_session,
                data=tilt_data,
            )

            eer_fractionation_file = None
            if self.data_collection_parameters.get("num_eer_frames"):
                response = capture_post(
                    base_url=str(environment.url.geturl()),
                    router_name="file_io_instrument.router",
                    function_name="write_eer_fractionation_file",
                    token=self._token,
                    visit_name=environment.visit,
                    session_id=environment.murfey_session,
                    data={
                        "num_frames": self.data_collection_parameters["num_eer_frames"],
                        "fractionation": self.data_collection_parameters[
                            "eer_fractionation"
                        ],
                        "dose_per_frame": environment.dose_per_frame or 0,
                        "fractionation_file_name": "eer_fractionation_tomo.txt",
                    },
                )
                eer_fractionation_file = response.json()["eer_fractionation_file"]
            preproc_data = {
                "path": str(file_transferred_to),
                "description": "",
                "image_number": environment.movies[file_transferred_to].movie_number,
                "pixel_size": self.data_collection_parameters.get(
                    "pixel_size_on_image", 0
                ),
                "dose_per_frame": environment.dose_per_frame or 0,
                "frame_count": self.data_collection_parameters.get("frame_count", 0),
                "tilt_axis": self.data_collection_parameters.get("tilt_axis", 85),
                "mc_binning": self.data_collection_parameters.get(
                    "motion_corr_binning", 1
                ),
                "gain_ref": environment.gain_ref,
                "voltage": self.data_collection_parameters.get("voltage", 300),
                "eer_fractionation_file": eer_fractionation_file,
                "tag": tilt_series,
                "group_tag": str(self._basepath),
            }
            capture_post(
                base_url=str(environment.url.geturl()),
                router_name="workflow.tomo_router",
                function_name="request_tomography_preprocessing",
                token=self._token,
                visit_name=environment.visit,
                session_id=environment.murfey_session,
                data=preproc_data,
            )

        return self._check_tilt_series(tilt_series)

    def _check_tilt_series(
        self,
        tilt_series: str,
    ) -> List[str]:
        newly_completed_series: List[str] = []
        mdoc_tilt_series_size = self._tilt_series_sizes.get(tilt_series, 0)
        if not self._tilt_series or not mdoc_tilt_series_size:
            logger.debug(f"Tilt series size not yet set for {tilt_series!r}")
            return newly_completed_series

        counted_tilts = len(self._tilt_series.get(tilt_series, []))
        tilt_series_size_check = counted_tilts >= mdoc_tilt_series_size
        if tilt_series_size_check and tilt_series not in self._completed_tilt_series:
            self._completed_tilt_series.append(tilt_series)
            newly_completed_series.append(tilt_series)
        else:
            logger.debug(
                f"{tilt_series!r} not complete yet. Counted {counted_tilts} tilts. "
                f"Expected number of tilts was {mdoc_tilt_series_size}"
            )
        return newly_completed_series

    def _add_tomo_tilt(
        self,
        file_path: Path,
        environment: MurfeyInstanceEnvironment | None = None,
        required_strings: List[str] | None = None,
    ) -> List[str]:
        required_strings = (
            ["fractions"] if required_strings is None else required_strings
        )
        if not any(r in file_path.name for r in required_strings):
            return []
        if environment:
            if tomo_version := environment.software_versions.get("tomo"):
                tilt_info_extraction = tomo_tilt_info.get(tomo_version)
                if not tilt_info_extraction:
                    raise ValueError(
                        f"Extraction routines for TFS Tomo version {tomo_version} unknown"
                    )
            else:
                tilt_info_extraction = tomo_tilt_info["5.7"]
        else:
            tilt_info_extraction = tomo_tilt_info["5.7"]
        return self._add_tilt(
            file_path,
            tilt_info_extraction,
            environment=environment,
            required_strings=required_strings,
        )

    def post_transfer(
        self,
        transferred_file: Path,
        environment: MurfeyInstanceEnvironment | None = None,
        **kwargs,
    ) -> List[str]:
        super().post_transfer(
            transferred_file=transferred_file,
            environment=environment,
            **kwargs,
        )

        data_suffixes = (".mrc", ".tiff", ".tif", ".eer")
        completed_tilts = []

        if "gain" not in transferred_file.name:
            if transferred_file.suffix in data_suffixes:
                if self._acquisition_software == "tomo":
                    if environment:
                        machine_config = get_machine_config_client(
                            str(environment.url.geturl()),
                            self._token,
                            instrument_name=environment.instrument_name,
                        )
                    else:
                        machine_config = {}
                    required_strings = (
                        machine_config.get("data_required_substrings", {})
                        .get("tomo", {})
                        .get(transferred_file.suffix, ["fractions"])
                    )
                    completed_tilts = self._add_tomo_tilt(
                        transferred_file,
                        environment=environment,
                        required_strings=kwargs.get("required_strings")
                        or required_strings,
                    )
                else:
                    logger.warning(f"Unknown data file {transferred_file}")
            if transferred_file.suffix == ".mdoc":
                with open(transferred_file, "r") as md:
                    tilt_series = transferred_file.stem
                    self._tilt_series_sizes[tilt_series] = get_num_blocks(md)
                if environment:
                    source = self._get_source(transferred_file, environment)
                    if source:
                        completed_tilts = self._check_tilt_series(tilt_series)

                    # Always update the tilt series length in the database after an mdoc
                    if environment.murfey_session is not None:
                        capture_post(
                            base_url=str(environment.url.geturl()),
                            router_name="workflow.tomo_router",
                            function_name="register_tilt_series_length",
                            token=self._token,
                            session_id=environment.murfey_session,
                            data={
                                "tags": [tilt_series],
                                "source": str(transferred_file.parent),
                                "tilt_series_lengths": [
                                    self._tilt_series_sizes[tilt_series]
                                ],
                            },
                        )

        if completed_tilts and environment:
            logger.info(
                f"The following tilt series are considered complete: {completed_tilts} "
                f"after {transferred_file}"
            )
            capture_post(
                base_url=str(environment.url.geturl()),
                router_name="workflow.tomo_router",
                function_name="register_completed_tilt_series",
                token=self._token,
                visit_name=environment.visit,
                session_id=environment.murfey_session,
                data={
                    "tags": completed_tilts,
                    "source": str(transferred_file.parent),
                    "tilt_series_lengths": [
                        len(self._tilt_series.get(ts, [])) for ts in completed_tilts
                    ],
                },
            )
        return completed_tilts

    def post_first_transfer(
        self,
        transferred_file: Path,
        environment: MurfeyInstanceEnvironment | None = None,
        **kwargs,
    ):
        self.post_transfer(transferred_file, environment=environment, **kwargs)

    def gather_metadata(
        self, metadata_file: Path, environment: MurfeyInstanceEnvironment | None = None
    ) -> OrderedDict:
        if metadata_file.suffix not in (".mdoc", ".xml"):
            raise ValueError(
                f"Tomography gather_metadata method expected xml or mdoc file not {metadata_file.name}"
            )
        try:
            if not metadata_file.is_file():
                logger.debug(f"Metadata file {metadata_file} not found")
                return OrderedDict({})
            if metadata_file.suffix == ".xml":
                with open(metadata_file, "r") as xml:
                    try:
                        for_parsing = xml.read()
                    except Exception:
                        logger.warning(
                            f"Failed to parse file {metadata_file}", exc_info=True
                        )
                        return OrderedDict({})
                    data = xmltodict.parse(for_parsing)
                try:
                    metadata: OrderedDict = OrderedDict({})
                    metadata["experiment_type"] = "tomography"
                    metadata["voltage"] = 300
                    metadata["image_size_x"] = data["Acquisition"]["Info"]["ImageSize"][
                        "Width"
                    ]
                    metadata["image_size_y"] = data["Acquisition"]["Info"]["ImageSize"][
                        "Height"
                    ]
                    metadata["pixel_size_on_image"] = float(
                        data["Acquisition"]["Info"]["SensorPixelSize"]["Height"]
                    )
                    metadata["motion_corr_binning"] = 1
                    metadata["gain_ref"] = None
                    metadata["dose_per_frame"] = (
                        environment.dose_per_frame if environment else None
                    )
                    metadata["source"] = str(self._basepath)
                except KeyError:
                    return OrderedDict({})
                return metadata
            with open(metadata_file, "r") as md:
                mdoc_data = get_global_data(md)
                num_blocks = get_num_blocks(md)
                md.seek(0)
                blocks = [get_block(md) for i in range(num_blocks)]
                mdoc_data_block = blocks[0]
            if not mdoc_data:
                return OrderedDict({})
            mdoc_metadata: OrderedDict = OrderedDict({})
            mdoc_metadata["experiment_type"] = "tomography"
            mdoc_metadata["voltage"] = float(mdoc_data["Voltage"])
            mdoc_metadata["frame_count"] = int(mdoc_data_block["NumSubFrames"])
            mdoc_metadata["tilt_axis"] = float(mdoc_data_block["RotationAngle"])
            mdoc_metadata["image_size_x"] = int(mdoc_data["ImageSize"][0])
            mdoc_metadata["image_size_y"] = int(mdoc_data["ImageSize"][1])
            mdoc_metadata["magnification"] = int(mdoc_data_block["Magnification"])
            superres_binning = int(mdoc_data_block["Binning"])
            binning_factor = 1
            if environment:
                server_config = capture_get(
                    base_url=str(environment.url.geturl()),
                    router_name="session_control.router",
                    function_name="machine_info_by_instrument",
                    token=self._token,
                    instrument_name=environment.instrument_name,
                ).json()
                if (
                    server_config.get("superres")
                    and superres_binning == 1
                    and not environment.superres
                ):
                    binning_factor = 2
                ps_from_mag = (
                    server_config.get("calibrations", {})
                    .get("magnification", {})
                    .get(mdoc_data_block["Magnification"])
                )
                if ps_from_mag:
                    mdoc_metadata["pixel_size_on_image"] = (
                        float(ps_from_mag) * 1e-10 / binning_factor
                    )
            if mdoc_metadata.get("pixel_size_on_image") is None:
                mdoc_metadata["pixel_size_on_image"] = (
                    float(mdoc_data["PixelSpacing"]) * 1e-10
                )
            mdoc_metadata["motion_corr_binning"] = binning_factor
            mdoc_metadata["gain_ref"] = environment.gain_ref if environment else None
            mdoc_metadata["dose_per_frame"] = (
                environment.dose_per_frame if environment else None
            )
            mdoc_metadata["source"] = str(self._basepath)
            mdoc_metadata["tag"] = str(self._basepath)
            mdoc_metadata["tilt_series_tag"] = metadata_file.stem
            mdoc_metadata["exposure_time"] = float(mdoc_data_block["ExposureTime"])
            slit_width = mdoc_data_block["FilterSlitAndLoss"][0]
            if slit_width.lower() != "nan":
                mdoc_metadata["slit_width"] = float(slit_width)
            mdoc_metadata["file_extension"] = (
                f".{mdoc_data_block['SubFramePath'].split('.')[-1]}"
            )
            mdoc_metadata["eer_fractionation"] = (
                environment.eer_fractionation if environment else None
            ) or 20

            data_file = mdoc_data_block["SubFramePath"].split("\\")[-1]
            if data_file.split(".")[-1] == "eer":
                mdoc_metadata["num_eer_frames"] = murfey.util.eer.num_frames(
                    metadata_file.parent / data_file
                )
                mdoc_metadata["frame_count"] = int(
                    int(mdoc_metadata["eer_fractionation"])
                    / int(mdoc_metadata["num_eer_frames"])
                )
        except Exception as e:
            logger.error(
                f"Exception encountered in metadata gathering: {str(e)}", exc_info=True
            )
            return OrderedDict({})

        return mdoc_metadata



import contextlib
import io
import json
from argparse import ArgumentParser
from pathlib import Path

import yaml
from fastapi.openapi.utils import get_openapi

import murfey
from murfey.cli import LineWrapHelpFormatter, PrettierDumper


def run():
    # Set up argument parser
    parser = ArgumentParser(
        description=(
            "Generates an OpenAPI schema of the chosen FastAPI server "
            "and outputs it as either a JSON or YAML file"
        ),
        formatter_class=LineWrapHelpFormatter,
    )
    parser.add_argument(
        "--target",
        "-t",
        default="server",
        help=(
            "The target FastAPI server to construct the OpenAPI schema for. \n"
            "OPTIONS: instrument-server | server \n"
            "DEFAULT: server"
        ),
    )
    parser.add_argument(
        "--output",
        "-o",
        default="yaml",
        help=(
            "Set the output format of the OpenAPI schema. \n"
            "OPTIONS: json | yaml \n"
            "DEFAULT: yaml"
        ),
    )
    parser.add_argument(
        "--to-file",
        "-f",
        default="",
        help=(
            "Alternative file path and file name to save the schema as. "
            "Can be a relative or absolute path. \n"
            "By default, the schema will be saved to 'murfey/utils/', "
            "and it will have the name 'openapi.json' or 'openapi.yaml'."
        ),
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        default=False,
        help="Shows additional steps when setting ",
    )
    args = parser.parse_args()

    # Load the relevant FastAPI app
    target = str(args.target).lower()

    # Silence output during import; only return genuine errors
    buffer = io.StringIO()
    with contextlib.redirect_stdout(buffer), contextlib.redirect_stderr(buffer):
        if target == "server":
            from murfey.server.main import app
        elif target == "instrument-server":
            from murfey.instrument_server.main import app
        else:
            raise ValueError(
                "Unexpected value for target server. It must be one of "
                "'instrument-server' or 'server'"
            )
    if args.debug:
        print(f"Imported FastAPI app for {target}")

    if not app.openapi_schema:
        schema = get_openapi(
            title=app.title,
            version=app.version,
            openapi_version=app.openapi_version,
            description=app.description,
            routes=app.routes,
        )
        if args.debug:
            print(f"Constructed OpenAPI schema for {target}")
    else:
        schema = app.openapi_schema
        if args.debug:
            print(f"Loaded OpenAPI schema for {target}")

    output = str(args.output).lower()
    if output not in ("json", "yaml"):
        raise ValueError(
            "Invalid file format selected. Output must be either 'json' or 'yaml'"
        )
    murfey_dir = Path(murfey.__path__[0])
    save_path = (
        murfey_dir / "util" / f"openapi-{target}.{output}"
        if not args.to_file
        else Path(args.to_file)
    )
    with open(save_path, "w") as f:
        if output == "json":
            json.dump(schema, f, indent=2)
        else:
            yaml.dump(
                schema,
                f,
                Dumper=PrettierDumper,
                default_flow_style=False,
                sort_keys=False,
                indent=2,
            )
    print(f"OpenAPI schema saved to {save_path}")
    exit()


if __name__ == "__main__":
    run()
